from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class DLA(Variable):
    value_type = float
    entity = Person
    label = u"Disability Living Allowance"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("DLA_M", period) + person("DLA_SC", period)


class DLA_M(Variable):
    value_type = float
    entity = Person
    label = u"Disability Living Allowance (mobility component)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("DLA_M_reported", period)


class DLA_SC(Variable):
    value_type = float
    entity = Person
    label = u"Disability Living Allowance (self-care)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("DLA_SC_reported", period)


class DLA_SC_middle_plus(Variable):
    value_type = bool
    entity = Person
    label = u"Receives at least DLA (self-care) middle rate"
    definition_period = YEAR

    def formula(person, period, parameters):
        DLA_SC = parameters(period).benefit.DLA.self_care
        return person("DLA_SC", period) >= DLA_SC.middle


class DLA_M_reported(Variable):
    value_type = float
    entity = Person
    label = u"Disability Living Allowance (mobility component) (reported)"
    definition_period = YEAR


class DLA_SC_reported(Variable):
    value_type = float
    entity = Person
    label = u"Disability Living Allowance (self-care) (reported)"
    definition_period = YEAR



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class AA(Variable):
    value_type = float
    entity = Person
    label = u"Attendance Allowance"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("AA_reported", period)


class AA_reported(Variable):
    value_type = float
    entity = Person
    label = u"Attendance Allowance (reported)"
    definition_period = YEAR



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class housing_benefit_reported(Variable):
    value_type = float
    entity = Person
    label = u"Housing Benefit (reported amount)"
    definition_period = YEAR


class housing_benefit_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether eligible for Housing Benefit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        social = benunit.any(benunit.members("in_social_housing", period))
        return social + benunit("LHA_eligible", period)


class would_claim_HB(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Would claim Housing Benefit"
    documentation = (
        "Whether this family would claim Housing Benefit if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return (
            random(benunit) < parameters(period).benefit.housing_benefit.takeup
        ) | benunit("claims_all_entitled_benefits", period)


class claims_HB(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Would claim Housing Benefit"
    documentation = (
        "Whether this family would claim Housing Benefit if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit("would_claim_HB", period) & benunit(
            "claims_legacy_benefits", period
        )


class housing_benefit_applicable_amount(Variable):
    value_type = float
    entity = BenUnit
    label = u"Applicable amount for Housing Benefit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        HB = parameters(period).benefit.housing_benefit
        PA = HB.allowances
        one_over_SP_age = benunit.any(benunit.members("is_SP_age", period))
        eldest_age = benunit("eldest_adult_age", period)
        u_18 = eldest_age < 18
        u_25 = eldest_age < 25
        o_25 = (eldest_age >= 25) * not_(one_over_SP_age)
        o_18 = (eldest_age >= 18) * not_(one_over_SP_age)
        single = benunit("is_single_person", period)
        couple = benunit("is_couple", period)
        lone_parent = benunit("is_lone_parent", period)
        personal_allowance = (
            single
            * (
                u_25 * PA.single.under_25
                + o_25 * PA.single.over_25
                + one_over_SP_age * PA.single.SP_age
            )
            + couple
            * (
                u_18 * PA.couple.both_under_18
                + o_18 * PA.couple.over_18
                + one_over_SP_age * PA.couple.SP_age
            )
            + lone_parent
            * (
                u_18 * PA.lone_parent.under_18
                + o_18 * PA.lone_parent.over_18
                + one_over_SP_age * PA.lone_parent.SP_age
            )
        ) * WEEKS_IN_YEAR
        premiums = benunit("benefits_premiums", period)
        return (personal_allowance + premiums) * benunit(
            "housing_benefit_eligible", period
        )


class housing_benefit_applicable_income(Variable):
    value_type = float
    entity = BenUnit
    label = u"Relevant income for Housing Benefit means test"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        means_test = parameters(period).benefit.housing_benefit.means_test
        BENUNIT_MEANS_TESTED_BENEFITS = [
            "child_benefit",
            "income_support",
            "JSA_income",
            "ESA_income",
        ]
        INCOME_COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "property_income",
            "pension_income",
        ]
        benefits = add(benunit, period, BENUNIT_MEANS_TESTED_BENEFITS)
        income = aggr(benunit, period, INCOME_COMPONENTS)
        tax = aggr(
            benunit,
            period,
            ["income_tax", "national_insurance"],
        )
        income += aggr(benunit, period, ["personal_benefits"])
        income += add(benunit, period, ["tax_credits"])
        income -= tax
        income -= aggr(benunit, period, ["pension_contributions"]) * 0.5
        income += benefits
        num_children = benunit.nb_persons(BenUnit.CHILD)
        max_childcare_amount = (
            num_children == 1
        ) * WTC.elements.childcare_1 * WEEKS_IN_YEAR + (
            num_children > 1
        ) * WTC.elements.childcare_2 * WEEKS_IN_YEAR
        childcare_element = min_(
            max_childcare_amount,
            benunit.sum(benunit.members("childcare_expenses", period)),
        )
        applicable_income = max_(
            0,
            income
            - benunit("is_single_person", period)
            * means_test.income_disregard_single
            * WEEKS_IN_YEAR
            - benunit("is_couple", period)
            * means_test.income_disregard_couple
            * WEEKS_IN_YEAR
            - benunit("is_lone_parent", period)
            * means_test.income_disregard_lone_parent
            * WEEKS_IN_YEAR
            - (
                (
                    benunit.sum(benunit.members("weekly_hours", period))
                    > means_test.worker_hours
                )
                + (
                    benunit("is_lone_parent", period)
                    * benunit.sum(benunit.members("weekly_hours", period))
                    > WTC.min_hours.lower
                )
            )
            * means_test.worker_income_disregard
            * WEEKS_IN_YEAR
            - childcare_element,
        )
        return applicable_income


class HB_individual_non_dep_deduction(Variable):
    value_type = float
    entity = Person
    label = u"Non-dependent deduction (individual)"
    definition_period = YEAR

    def formula(person, period, parameters):
        not_rent_liable = person.benunit("benunit_rent", period) == 0
        over_21 = person("age", period) >= 21
        deduction_scale = parameters(
            period
        ).benefit.housing_benefit.deductions.non_dep_deduction
        weekly_income = person("total_income", period)
        deduction = deduction_scale.calc(weekly_income)
        return deduction * over_21 * not_rent_liable * MONTHS_IN_YEAR


class HB_non_dep_deductions(Variable):
    value_type = float
    entity = BenUnit
    label = u"Non-dependent deductions"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        non_dep_deductions_in_hh = benunit.max(
            benunit.members.household.sum(
                benunit.members("HB_individual_non_dep_deduction", period)
            )
        )
        non_dep_deductions_in_bu = benunit.sum(
            benunit.members("HB_individual_non_dep_deduction", period)
        )
        return non_dep_deductions_in_hh - non_dep_deductions_in_bu


class housing_benefit(Variable):
    value_type = float
    entity = BenUnit
    label = u"Housing Benefit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        rent = benunit("benunit_rent", period)
        LHA = benunit("LHA_eligible", period.this_year)
        applicable_amount = benunit(
            "housing_benefit_applicable_amount", period
        )
        income = benunit("housing_benefit_applicable_income", period)
        withdrawal_rate = parameters(
            period
        ).benefit.housing_benefit.means_test.withdrawal_rate
        final_amount = max_(
            0, rent - max_(0, income - applicable_amount) * withdrawal_rate
        )
        amount = where(
            LHA, min_(final_amount, benunit("LHA_cap", period)), final_amount
        )
        other_capped_benefits = add(
            benunit,
            period,
            [
                "child_benefit",
                "child_tax_credit",
                "JSA_income",
                "income_support",
                "ESA_income",
            ],
        ) + aggr(
            benunit,
            period,
            ["JSA_contrib", "incapacity_benefit", "ESA_contrib", "SDA"],
        )
        amount = max_(0, amount - benunit("HB_non_dep_deductions", period))
        final_amount = min_(
            amount * benunit("claims_HB", period),
            benunit("benefit_cap", period) - other_capped_benefits,
        )
        return max_(0, final_amount)



from openfisca_uk import BASELINE_VARIABLES
from openfisca_core.model_api import Enum


def test_variable_names_match_return_types():
    exceptions = []
    for var_name in BASELINE_VARIABLES:
        variable = BASELINE_VARIABLES[var_name]
        try:
            if var_name[:4] == "num_":
                assert variable.value_type == int
            elif var_name[:3] == "is_":
                assert variable.value_type == bool
            else:
                assert variable.value_type in (Enum, float, bool, str, int)
        except Exception as e:
            exceptions += [f"{var_name} returns {variable.value_type}"]
    print(
        f"Total variables: {len(BASELINE_VARIABLES)}, variable name violations: {len(exceptions)}."
    )
    for exception in exceptions:
        print(f"\t{exception}")
    if len(exceptions) > 0:
        raise Exception("Some variable names do not match their output types.")


def test_labels_exist():
    messages = []
    for var_name in BASELINE_VARIABLES:
        if not hasattr(BASELINE_VARIABLES[var_name], "label"):
            messages += [f"{var_name} has no label set"]
        elif BASELINE_VARIABLES[var_name].label == "label":
            messages += [f"{var_name} has the default label"]
    if len(messages) > 0:
        print("\n".join(messages))
        raise Exception("Some variables do not have labels.")



# -*- coding: utf-8 -*-

# This file defines the entities needed by our legislation.
from openfisca_core.entities import build_entity

State = build_entity(
    key="state",
    plural="states",
    label="State",
    roles=[
        {
            "key": "citizen",
            "plural": "citizens",
            "label": "Citizen",
            "doc": "A person who is a citizen of a country.",
        }
    ],
)

Household = build_entity(
    key="household",
    plural="households",
    label=u"Household",
    doc=u"Each household may contain multiple families, or benefit units.",
    roles=[
        {
            "key": "adult",
            "plural": "adults",
            "label": u"Adult",
            "doc": u"The adults in the household.",
        },
        {
            "key": "child",
            "plural": "children",
            "label": u"Child",
            "doc": u"Dependent children in the household.",
        },
    ],
    containing_entities=["state"],
)

BenUnit = build_entity(
    key="benunit",
    plural="benunits",
    label=u"Benefit unit",
    doc=u"A family is also known as a benefit unit, and contains parents and dependent children.",
    roles=[
        {
            "key": "adult",
            "plural": "adults",
            "label": u"Adult",
            "max": 2,
            "doc": u"The one or two adults in the benefit unit.",
        },
        {
            "key": "child",
            "plural": "children",
            "label": u"Child",
            "doc": u"Dependent children.",
        },
    ],
    containing_entities=["household", "state"],
)

Person = build_entity(
    key="person",
    plural="people",
    label="Person",
    doc=u"An individual. The minimal legal entity on which a legislation might be applied.",
    is_person=True,
)

entities = [State, Household, BenUnit, Person]



# -*- coding: utf-8 -*-

import json
import os


DIR_PATH = os.path.dirname(os.path.abspath(__file__))


def parse(file_name):
    file_path = os.path.join(DIR_PATH, file_name)
    with open(file_path, "r") as file:
        return json.loads(file.read())



from openfisca_uk_data import FRS
from openfisca_uk import Microsimulation
from openfisca_tools.testing import generate_tests


test_parameter = generate_tests(Microsimulation(dataset=FRS, year=2019))



from openfisca_uk.variables.demographic.household import TenureType
from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class benunit_id(Variable):
    value_type = int
    entity = BenUnit
    label = u"ID for the family"
    definition_period = YEAR


class families(Variable):
    value_type = float
    entity = BenUnit
    label = u"Variable holding families"
    definition_period = YEAR
    default_value = 1


class benunit_weight(Variable):
    value_type = float
    entity = BenUnit
    label = u"Weight factor for the benefit unit"
    definition_period = YEAR


class is_married(Variable):
    value_type = bool
    entity = BenUnit
    label = "Married"
    documentation = "Whether the benefit unit adults are married to each other"
    definition_period = YEAR


class FamilyType(Enum):
    SINGLE = u"Single, with no children"
    COUPLE_NO_CHILDREN = u"Couple, with no children"
    LONE_PARENT = u"Lone parent, with children"
    COUPLE_WITH_CHILDREN = u"Couple, with children"


class RelationType(Enum):
    SINGLE = u"Single"
    COUPLE = u"Couple"


class family_type(Variable):
    value_type = Enum
    entity = BenUnit
    default_value = FamilyType.SINGLE
    possible_values = FamilyType
    label = u"Family composition"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        two_adults = benunit.sum(benunit.members("is_adult", period)) == 2
        has_children = benunit.sum(benunit.members("is_child", period)) > 0
        is_single = not_(two_adults) * not_(has_children)
        is_couple = two_adults * not_(has_children)
        is_lone = not_(two_adults) * has_children
        is_full = two_adults * has_children
        return select([is_single, is_couple, is_lone, is_full], FamilyType)


class relation_type(Variable):
    value_type = Enum
    entity = BenUnit
    default_value = RelationType.SINGLE
    possible_values = RelationType
    label = u"Whether single or a couple"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return where(
            benunit.sum(benunit.members("is_adult", period)) == 1,
            RelationType.SINGLE,
            RelationType.COUPLE,
        )


class eldest_adult_age(Variable):
    value_type = float
    entity = BenUnit
    label = u"Eldest adult age"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.max(
            benunit.members("age", period.this_year), role=BenUnit.ADULT
        )


class youngest_adult_age(Variable):
    value_type = float
    entity = BenUnit
    label = u"Eldest adult age"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.min(
            benunit.members("age", period.this_year), role=BenUnit.ADULT
        )


class eldest_child_age(Variable):
    value_type = float
    entity = BenUnit
    label = u"Eldest adult age"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.max(
            benunit.members("age", period.this_year), role=BenUnit.CHILD
        )


class youngest_child_age(Variable):
    value_type = float
    entity = BenUnit
    label = u"Eldest adult age"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.min(
            benunit.members("age", period.this_year), role=BenUnit.CHILD
        )


class num_children(Variable):
    value_type = int
    entity = BenUnit
    label = u"The number of children in the family"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.sum(benunit.members("is_child", period))


class num_adults(Variable):
    value_type = int
    entity = BenUnit
    label = u"The number of adults in the family"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.sum(benunit.members("is_adult", period))


class benunit_tenure_type(Variable):
    value_type = Enum
    possible_values = TenureType
    default_value = TenureType.RENT_PRIVATELY
    entity = BenUnit
    label = u"Tenure type of the family's household"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        tenure = benunit.value_from_first_person(
            benunit.members.household("tenure_type", period)
        )
        return tenure


class benunit_is_renting(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether this family is renting"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        tenure = benunit("benunit_tenure_type", period)
        return np.isin(
            tenure,
            (
                TenureType.RENT_PRIVATELY,
                TenureType.RENT_FROM_COUNCIL,
                TenureType.RENT_FROM_HA,
            ),
        )


class benunit_random_number(Variable):
    label = "Randomness"
    documentation = (
        "A random number between zero and one generated for this benefit unit"
    )
    entity = BenUnit
    definition_period = YEAR
    value_type = float

    def formula(benunit, period):
        return random(benunit)



from openfisca_uk.model_api import *


class stamp_duty_on_residential_property(Variable):
    label = "Stamp Duty on residential property"
    documentation = (
        "Tax charge from purchase or rental of residential property"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        stamp_duty = parameters(period).hmrc.stamp_duty
        # Tax on main-home purchases
        price_limit = stamp_duty.residential.purchase.main.first.max
        price = household("main_residential_property_purchased", period)
        residential_purchase_qualifies_as_first_buy = household(
            "main_residential_property_purchased_is_first_home", period
        ) & (price < price_limit)
        main_residential_purchase_tax = where(
            residential_purchase_qualifies_as_first_buy,
            stamp_duty.residential.purchase.main.first.rate.calc(price),
            stamp_duty.residential.purchase.main.subsequent.calc(price),
        )
        # Tax on second-home purchases
        second_home_price = household(
            "additional_residential_property_purchased", period
        )
        price = where(
            second_home_price < stamp_duty.residential.purchase.additional.min,
            0,
            second_home_price,
        )
        additional_residential_purchase_tax = (
            stamp_duty.residential.purchase.additional.rate.calc(price)
        )
        # Tax on residential rents
        cumulative_rent = household("cumulative_residential_rent", period)
        rent = household("rent", period)
        residential_rent_tax = stamp_duty.residential.rent.calc(
            cumulative_rent + rent
        ) - stamp_duty.residential.rent.calc(cumulative_rent)
        return household("stamp_duty_liable", period) * (
            main_residential_purchase_tax
            + additional_residential_purchase_tax
            + residential_rent_tax
        )


class stamp_duty_on_non_residential_property(Variable):
    label = "Stamp Duty on non-residential property"
    documentation = (
        "Tax charge from purchase or rental of non-residential property"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        stamp_duty = parameters(period).hmrc.stamp_duty
        # Tax on non-residential purchases
        price = household("non_residential_property_purchased", period)
        non_residential_purchase_tax = (
            stamp_duty.non_residential.purchase.calc(price)
        )
        # Tax on non-residential rents
        cumulative_rent = household("cumulative_non_residential_rent", period)
        rent = household("non_residential_rent", period)
        non_residential_rent_tax = stamp_duty.non_residential.rent.calc(
            cumulative_rent + rent
        ) - stamp_duty.non_residential.rent.calc(cumulative_rent)
        return household("stamp_duty_liable", period) * (
            non_residential_purchase_tax + non_residential_rent_tax
        )


class stamp_duty_liable(Variable):
    label = "Liable for Stamp Duty"
    documentation = "Whether the household is liable for Stamp Duty Land Tax"
    entity = Household
    definition_period = YEAR
    value_type = bool
    unit = "currency-GBP"

    def formula(household, period):
        country = household("country", period)
        countries = country.possible_values
        return np.isin(
            country.decode_to_str(),
            [countries.ENGLAND.name, countries.NORTHERN_IRELAND.name],
        )


class corporate_stamp_duty_revenue(Variable):
    label = "Stamp Duty (corporations)"
    documentation = "Stamp Duty paid by corporations"
    entity = State
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(state, period, parameters):
        sd = parameters(period).hmrc.stamp_duty.statistics
        in_force = not parameters(period).hmrc.stamp_duty.abolition
        return in_force * (
            sd.residential.corporate.revenue
            + sd.non_residential.corporate.revenue
        )


class corporate_stamp_duty_revenue_change(Variable):
    label = "Change in corporate Stamp Duty liabilities"
    documentation = (
        "Tota increase in tax revenues from corporations for Stamp Duty"
    )
    entity = State
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(state, period, parameters):
        sd = parameters(period).hmrc.stamp_duty.statistics
        baseline_sd = (
            sd.residential.corporate.revenue
            + sd.non_residential.corporate.revenue
        )
        return state("corporate_stamp_duty_revenue", period) - baseline_sd


class stamp_duty(Variable):
    label = "Stamp Duty Land Tax"
    documentation = "Total tax liability for Stamp Duty Land Tax"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period):
        return household(
            "stamp_duty_on_residential_property", period
        ) + household("stamp_duty_on_non_residential_property", period)


class expected_stamp_duty(Variable):
    label = "Stamp Duty (expected)"
    documentation = "Expected value of Stamp Duty taxes"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        in_force = not parameters(period).hmrc.stamp_duty.abolition
        return (
            in_force
            * household("property_sale_rate", period)
            * household("stamp_duty", period)
        )



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class income_support_reported(Variable):
    value_type = float
    entity = Person
    label = u"Income Support (reported amount)"
    definition_period = YEAR


class would_claim_IS(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Would claim Income Support"
    documentation = (
        "Whether this family would claim Income Support if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return (
            random(benunit) <= parameters(period).benefit.income_support.takeup
        ) + benunit("claims_all_entitled_benefits", period)


class claims_IS(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether this family is imputed to claim Income Support"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit("would_claim_IS", period) & benunit(
            "claims_legacy_benefits", period
        )


class income_support_applicable_income(Variable):
    value_type = float
    entity = BenUnit
    label = u"Relevant income for Income Support means test"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        IS = parameters(period).benefit.income_support
        INCOME_COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "property_income",
            "pension_income",
        ]
        income = aggr(benunit, period, INCOME_COMPONENTS)
        tax = aggr(
            benunit,
            period,
            ["income_tax", "national_insurance"],
        )
        income += aggr(benunit, period, ["social_security_income"])
        income -= tax
        income -= aggr(benunit, period, ["pension_contributions"]) * 0.5
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        income = max_(
            0,
            income
            - (family_type == families.SINGLE)
            * IS.means_test.income_disregard_single
            * WEEKS_IN_YEAR
            - benunit("is_couple", period)
            * IS.means_test.income_disregard_couple
            * WEEKS_IN_YEAR
            - (family_type == families.LONE_PARENT)
            * IS.means_test.income_disregard_lone_parent
            * WEEKS_IN_YEAR,
        )
        return income


class income_support_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether eligible for Income Support"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        lone_parent = benunit("is_lone_parent", period) * (
            benunit("youngest_child_age", period) <= 5
        )
        QUALIFYING_COMPONENTS = ["is_carer_for_benefits"]
        eligible = (
            aggr(benunit, period, QUALIFYING_COMPONENTS) + lone_parent
        ) > 0
        under_SP_age = benunit.any(benunit.members("is_SP_age", period)) == 0
        eligible *= under_SP_age
        return not_(benunit("ESA_income", period) > 0) * eligible


class income_support_applicable_amount(Variable):
    value_type = float
    entity = BenUnit
    label = u"Applicable amount of Income Support"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        IS = parameters(period).benefit.income_support
        amounts = IS.amounts
        younger_age = benunit("youngest_adult_age", period)
        older_age = benunit("eldest_adult_age", period)
        children = benunit.sum(benunit.members("is_child", period)) > 0
        single = benunit("is_single", period)
        single_under_25 = single * not_(children) * (younger_age < 25)
        single_over_25 = single * not_(children) * (younger_age >= 25)
        lone_young = single * children * (younger_age < 18)
        lone_old = single * children * (younger_age >= 18)
        couple_young = not_(single) * (older_age < 18)
        couple_mixed = not_(single) * (older_age >= 18) * (younger_age < 18)
        couple_old = not_(single) * (younger_age >= 18)
        personal_allowance = (
            select(
                [
                    single_under_25,
                    single_over_25,
                    lone_young,
                    lone_old,
                    couple_young,
                    couple_mixed,
                    couple_old,
                ],
                [
                    amounts.amount_16_24,
                    amounts.amount_over_25,
                    amounts.amount_lone_16_17,
                    amounts.amount_lone_over_18,
                    amounts.amount_couples_16_17,
                    amounts.amount_couples_age_gap,
                    amounts.amount_couples_over_18,
                ],
            )
            * WEEKS_IN_YEAR
        )
        premiums = benunit("benefits_premiums", period)
        return (
            (personal_allowance + premiums)
            * benunit("income_support_eligible", period)
            * benunit("claims_IS", period)
        )


class income_support(Variable):
    value_type = float
    entity = BenUnit
    label = u"Income Support"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        amount = benunit("income_support_applicable_amount", period)
        income = benunit("income_support_applicable_income", period)
        return max_(
            0,
            (amount - income),
        )



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class receives_carers_allowance(Variable):
    value_type = bool
    entity = Person
    label = "Receives Carer's Allowance"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("carers_allowance", period) > 0


class carers_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Carer's Allowance"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("carers_allowance_reported", period)


class carers_allowance_reported(Variable):
    value_type = float
    entity = Person
    label = u"Carer's Allowance (reported)"
    definition_period = YEAR



from openfisca_uk.model_api import *


class carbon_consumption(Variable):
    entity = Household
    label = "Carbon consumption"
    documentation = "Estimated total carbon footprint of the household"
    unit = "tonne CO2"
    definition_period = YEAR
    value_type = float

    def formula(household, period, parameters):
        CONSUMPTION_VARIABLES = [
            "food_and_non_alcoholic_beverages_consumption",
            "alcohol_and_tobacco_consumption",
            "clothing_and_footwear_consumption",
            "housing_water_and_electricity_consumption",
            "household_furnishings_consumption",
            "health_consumption",
            "transport_consumption",
            "communication_consumption",
            "recreation_consumption",
            "education_consumption",
            "restaurants_and_hotels_consumption",
            "miscellaneous_consumption",
        ]
        spending_by_sector = list(
            map(lambda var: household(var, period), CONSUMPTION_VARIABLES)
        )
        household_weight = household("household_weight", period)
        aggregate_spending_by_sector = list(
            map(
                lambda values: (values * household_weight).sum(),
                spending_by_sector,
            )
        )
        carbon = parameters(period).consumption.carbon
        aggregate_emissions_by_sector = [
            carbon.emissions[category.replace("_consumption", "")]
            for category in CONSUMPTION_VARIABLES
        ]
        carbon_intensity_by_sector = [
            emissions / spending if spending > 0 else carbon.intensity[name]
            for emissions, spending, name in zip(
                aggregate_emissions_by_sector,
                aggregate_spending_by_sector,
                CONSUMPTION_VARIABLES,
            )
        ]
        return sum(
            [
                spending * carbon_intensity
                for spending, carbon_intensity in zip(
                    spending_by_sector, carbon_intensity_by_sector
                )
            ]
        )



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class ESA_contrib(Variable):
    value_type = float
    entity = Person
    label = u"ESA (contribution-based)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("ESA_contrib_reported", period)


class ESA_contrib_reported(Variable):
    value_type = float
    entity = Person
    label = u"Employment and Support Allowance (contribution-based) (reported)"
    definition_period = YEAR



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class benefit_cap(Variable):
    value_type = float
    entity = BenUnit
    label = u"Benefit cap for the family"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        children = benunit("num_children", period) > 0
        region = benunit.value_from_first_person(
            benunit.members.household("region", period)
        )
        regions = benunit.members.household("region", period).possible_values
        in_london = region == regions.LONDON
        cap = parameters(period).benefit.benefit_cap
        rate = (
            select(
                [
                    children * in_london,
                    children * not_(in_london),
                    not_(children) * in_london,
                    not_(children) * not_(in_london),
                ],
                [
                    cap.london_children,
                    cap.has_children,
                    cap.london_no_children,
                    cap.no_children,
                ],
            )
            * WEEKS_IN_YEAR
        )
        exempt = benunit("is_benefit_cap_exempt", period)
        cap = where(exempt, np.inf * np.ones_like(children), rate)
        return cap


class is_benefit_cap_exempt(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether exempt from the benefits cap"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        QUAL_PERSONAL_BENEFITS = [
            "carers_allowance",
            "DLA_SC",
            "DLA_M",
            "ESA_contrib",
        ]
        qualifying = (
            add(benunit, period, ["working_tax_credit"])
            + add(benunit, period, ["ESA_income"])
            + aggr(benunit, period, QUAL_PERSONAL_BENEFITS)
            > 0
        )
        return qualifying



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *

"""
These variables calculate the tax reliefs that a taxpayer is eligible for, following the Income Tax Act (2007) s. 23.
Tax reliefs are amounts to be deducted from the assessed total income of a person.
The section detailing some tax reliefs applicable is section 24 of the Act, but others are described in the 2003 and 2005 Acts as deductions from the respective components.
Not all reliefs are calculated here - only the major ones.
"""

# Employment income


class taxable_employment_income(Variable):
    value_type = float
    entity = Person
    label = u"Net taxable earnings"
    definition_period = YEAR
    reference = "Income Tax (Earnings and Pensions) Act 2003 s. 11"

    def formula(person, period, parameters):
        taxable_earnings = person("employment_income", period)
        deductions = add(
            person, period, ["employment_deductions", "pension_contributions"]
        )
        benefits = person("employment_benefits", period)
        return max_(0, taxable_earnings + benefits - deductions)


class employment_benefits(Variable):
    value_type = float
    entity = Person
    label = u"Employment benefits"
    definition_period = YEAR

    def formula(person, period, parameters):
        return add(person, period, ["SSP", "SMP"])


class SMP(Variable):
    value_type = float
    entity = Person
    label = u"SMP"
    definition_period = YEAR


class SSP(Variable):
    value_type = float
    entity = Person
    label = u"Statutory Sick Pay"
    definition_period = YEAR


class employment_deductions(Variable):
    value_type = float
    entity = Person
    label = u"Deductions from employment income"
    definition_period = YEAR
    reference = "Income Tax Act (Earnings and Pensions) Act 2003 s. 327"

    def formula(person, period, parameters):
        DEDUCTIONS = ["employment_expenses"]
        return add(person, period, DEDUCTIONS)


class employment_expenses(Variable):
    value_type = float
    entity = Person
    label = (
        u"Cost of expenses necessarily incurred and reimbursed by employment"
    )
    definition_period = YEAR
    reference = "Income Tax Act (Earnings and Pensions) Act 2003 s. 333"


class pension_contributions(Variable):
    value_type = float
    entity = Person
    label = u"Amount contributed to registered pension schemes paid by the individual (not the employer)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return add(
            person,
            period,
            (
                "private_pension_contributions",
                "occupational_pension_contributions",
            ),
        )


class pension_contributions_relief(Variable):
    value_type = float
    entity = Person
    label = u"Reduction in taxable income from pension contributions"
    definition_period = YEAR
    reference = "Finance Act 2004 s. 188-194"

    def formula_2004_07_22(person, period, parameters):
        contributions = person("pension_contributions", period)
        pay = add(
            person, period, ["employment_income", "self_employment_income"]
        )
        under_75 = person("age", period) < 75
        basic_amount = parameters(
            period
        ).tax.income_tax.reliefs.pension_contribution.basic_amount
        tax_relief = min_(pay, max_(basic_amount, contributions)) * under_75
        capped_relief = min_(
            tax_relief, person("pension_annual_allowance", period)
        )
        return capped_relief


# Savings interest income


class taxable_savings_interest_income(Variable):
    value_type = float
    entity = Person
    label = u"Amount of savings interest which is taxable"
    definition_period = YEAR
    reference = "Income Tax Act (Trading and Other Income) 2005 s. 369"

    def formula(person, period, parameters):
        total_interest = person("savings_interest_income", period)
        exempt_interest = person("tax_free_savings_income", period)
        return max_(0, total_interest - exempt_interest)


class tax_free_savings_income(Variable):
    value_type = float
    entity = Person
    label = u"Income from savings in tax-free accounts"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("ISA_interest_income", period)


class ISA_interest_income(Variable):
    value_type = float
    entity = Person
    label = u"Amount received in interest from Individual Savings Accounts"
    definition_period = YEAR


# Trading income


class trading_loss(Variable):
    value_type = float
    entity = Person
    label = u"Loss from trading in the current year."
    definition_period = YEAR


class capital_allowances(Variable):
    value_type = float
    entity = Person
    label = u"Full relief from capital expenditure allowances"
    definition_period = YEAR
    reference = "Capital Allowances Act 2001 s. 1"


class loss_relief(Variable):
    value_type = float
    entity = Person
    label = u"Tax relief from trading losses"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 59"
    documentation = u"Can be set against general income."

    def formula(person, period, parameters):
        current_loss = person("trading_loss", period)
        previous_loss = person("trading_loss", period.last_year)
        return current_loss + previous_loss


# Pension income


class taxable_pension_income(Variable):
    value_type = float
    entity = Person
    label = u"Amount of pension income that is taxable"
    definition_period = YEAR
    reference = "Income Tax (Earnings and Pensions) Act 2003 s. 567"

    def formula(person, period, parameters):
        return person("pension_income", period)


# Social security income


class taxable_social_security_income(Variable):
    value_type = float
    entity = Person
    label = u"Amount of social security income that is taxable"
    definition_period = YEAR
    reference = "Income Tax (Earnings and Pensions) Act 2003 s. 658"

    def formula(person, period, parameters):
        return person("social_security_income", period)


# Trading income


class taxable_self_employment_income(Variable):
    value_type = float
    entity = Person
    label = u"Amount of trading income that is taxable"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 5"

    def formula(person, period, parameters):
        DEDUCTIONS = ["loss_relief", "capital_allowances", "trading_allowance"]
        return max_(
            0,
            person("self_employment_income", period)
            - add(person, period, DEDUCTIONS),
        )
        return amount


# Property income


class taxable_property_income(Variable):
    value_type = float
    entity = Person
    label = u"Amount of property income that is taxable"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 268"

    def formula(person, period, parameters):
        return max_(
            0,
            person("property_income", period)
            - person("property_allowance", period),
        )


# Dividend income


class deficiency_relief(Variable):
    value_type = float
    entity = Person
    label = u"Deficiency relief"
    definition_period = YEAR


class taxable_dividend_income(Variable):
    value_type = float
    entity = Person
    label = u"Amount of dividend income that is taxable"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 383"

    def formula(person, period, parameters):
        return max_(
            0,
            person("dividend_income", period)
            - person("deficiency_relief", period),
        )


# Miscellaneous income


class taxable_miscellaneous_income(Variable):
    value_type = float
    entity = Person
    label = u"Amount of miscellaneous income that is taxable"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 574"

    def formula(person, period, parameters):
        return person("miscellaneous_income", period)


class total_income(Variable):
    value_type = float
    entity = Person
    label = u"Taxable income after tax reliefs and before allowances"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 23"

    def formula(person, period, parameters):
        COMPONENTS = [
            "employment_income",
            "pension_income",
            "social_security_income",
            "self_employment_income",
            "property_income",
            "savings_interest_income",
            "dividend_income",
            "miscellaneous_income",
        ]
        return add(person, period, COMPONENTS)


class adjusted_net_income(Variable):
    value_type = float
    entity = Person
    label = u"Taxable income after tax reliefs and before allowances"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 23"

    def formula(person, period, parameters):
        COMPONENTS = [
            "taxable_employment_income",
            "taxable_pension_income",
            "taxable_social_security_income",
            "taxable_self_employment_income",
            "taxable_property_income",
            "taxable_savings_interest_income",
            "taxable_dividend_income",
            "taxable_miscellaneous_income",
        ]
        return max_(
            0,
            add(person, period, COMPONENTS),
        )



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *
from openfisca_uk.variables.finance.tax.income_tax.liability import TaxBand

"""
This file calculates the allowances to which taxpayers are entitled. This follows step 3 of the Income Tax Act 2007 s. 23.
"""


class personal_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Personal Allowance for the year"
    unit = "currency-GBP"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 35"

    def formula(person, period, parameters):
        PA = parameters(period).tax.income_tax.allowances.personal_allowance
        ANI = person("adjusted_net_income", period)
        excess = max_(0, ANI - PA.maximum_ANI)
        reduction = excess * PA.reduction_rate
        amount = max_(0, PA.amount - reduction)
        return amount


class blind_persons_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Blind Person's Allowance for the year (not simulated)"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 38"


class unused_personal_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Unused personal allowance"
    definition_period = YEAR

    def formula(person, period, parameters):
        return max_(
            person("personal_allowance", period)
            - person("adjusted_net_income", period),
            0,
        )


class marriage_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Marriage Allowance for the year (a tax-reducer, rather than an allowance or tax relief)"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 55"

    def formula(person, period, parameters):
        tax_band = person("tax_band", period)
        bands = tax_band.possible_values
        under_PA = person("tax_band", period) == bands.NONE
        band = person("tax_band", period)
        # Marriage Allowance is eligible for couples with one in the basic rate bracket and
        # one not in a tax bracket. For those paying Scottish rates, the taxpayer can be in
        # the starter or intermediate rate brackets.
        meets_band_eligibility = np.isin(
            band, (bands.STARTER, bands.BASIC, bands.INTERMEDIATE)
        )
        marital = person("marital_status", period)
        married = marital == marital.possible_values.MARRIED
        transferable_allowance = person.benunit.sum(
            married
            * under_PA
            * min_(
                person("unused_personal_allowance", period),
                0.1 * person("personal_allowance", period),
            )
        )
        return married * meets_band_eligibility * transferable_allowance


class married_couples_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Married Couples' allowance for the year"
    definition_period = YEAR


class married_couples_allowance_deduction(Variable):
    value_type = float
    entity = Person
    label = u"Deduction from Married Couples' allowance for the year"
    definition_period = YEAR

    def formula(person, period, parameters):
        rate = parameters(
            period
        ).tax.income_tax.allowances.married_couples_allowance.deduction_rate
        return person("married_couples_allowance", period) * rate


class pension_annual_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Annual Allowance for pension contributions"
    definition_period = YEAR

    def formula(person, period, parameters):
        allowance = parameters(
            period
        ).tax.income_tax.allowances.annual_allowance
        ANI = person("adjusted_net_income", period)
        reduction = max_(0, ANI - allowance.taper) * allowance.reduction_rate
        amount = max_(allowance.minimum, allowance.default - reduction)
        return amount


class trading_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Trading Allowance for the year"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 783AF"

    def formula(person, period, parameters):
        return parameters(period).tax.income_tax.allowances.trading_allowance


class trading_allowance_deduction(Variable):
    value_type = float
    entity = Person
    label = u"Deduction applied by the trading allowance"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 783AF"

    def formula(person, period, parameters):
        amount = min_(
            person("trading_allowance", period),
            person("self_employment_income", period),
        )
        return amount


class property_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Property Allowance for the year"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 783BF"

    def formula(person, period, parameters):
        return parameters(period).tax.income_tax.allowances.property_allowance


class property_allowance_deduction(Variable):
    value_type = float
    entity = Person
    label = u"Deduction applied by the property allowance"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 783AF"

    def formula(person, period, parameters):
        amount = min_(
            person("property_income", period),
            person("property_allowance", period),
        )
        return amount


class savings_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Savings Allowance for the year"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 12B"

    def formula(person, period, parameters):
        tax_band = person("tax_band", period)
        tax_bands = tax_band.possible_values
        amounts = parameters(
            period
        ).tax.income_tax.allowances.personal_savings_allowance
        allowance = select(
            [
                tax_band == tax_bands.ADDITIONAL,
                tax_band == tax_bands.HIGHER,
                tax_band == tax_bands.BASIC,
                tax_band == tax_bands.NONE,
            ],
            [amounts.additional, amounts.higher, amounts.basic, amounts.basic],
        )
        return allowance


class dividend_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Dividend allowance for the person"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 13A"

    def formula(person, period, parameters):
        amount = parameters(
            period
        ).tax.income_tax.allowances.dividend_allowance
        return amount


class gift_aid(Variable):
    value_type = float
    entity = Person
    label = u"Expenditure under Gift Aid"
    definition_period = YEAR


class covenanted_payments(Variable):
    value_type = float
    entity = Person
    label = u"Covenanted payments to charities"
    definition_period = YEAR


class charitable_investment_gifts(Variable):
    value_type = float
    entity = Person
    label = u"Gifts of qualifying investment or property to charities"
    definition_period = YEAR


class other_deductions(Variable):
    value_type = float
    entity = Person
    label = u"All other tax deductions"
    definition_period = YEAR


class allowances(Variable):
    value_type = float
    entity = Person
    label = u"Allowances applicable to adjusted net income"
    definition_period = YEAR

    def formula(person, period, parameters):
        ALLOWANCES = [
            "personal_allowance",
            "blind_persons_allowance",
            "gift_aid",
            "covenanted_payments",
            "charitable_investment_gifts",
            "other_deductions",
        ]
        allowance = add(person, period, ALLOWANCES)
        return allowance



"""
This module tests the FRS dataset produced by openfisca-uk-data against UKMOD - checking that the distributions are similar.
"""

import pandas as pd
from openfisca_uk_data import FRS, UKMODOutput
from openfisca_uk import Microsimulation, REPO
import pytest
from microdf import MicroDataFrame
from openfisca_uk.reforms.presets.current_date import use_current_parameters
from openfisca_uk.tools.parameters import backdate_parameters
import yaml

TEST_YEAR = 2018
# Variable pairs to check for similarity
with open(
    REPO / "tests" / "microsimulation" / "frs" / "variable_ukmod_map.yml"
) as f:
    metadata = yaml.safe_load(f)

MAX_REL_ERROR = 0.05
MAX_MEAN_REL_ERROR = 0.05
MIN_NONZERO_AGREEMENT = 0.99

if TEST_YEAR not in UKMODOutput.years:
    raise FileNotFoundError("UKMOD FRS needed to run tests against.")
if TEST_YEAR not in FRS.years:
    raise FileNotFoundError("FRS needed to run tests.")

baseline = Microsimulation(
    dataset=FRS,
    year=TEST_YEAR,
)
ukmod = UKMODOutput.load(TEST_YEAR, "person")
ukmod_hh = ukmod.groupby("household_id").sum()
ukmod = MicroDataFrame(ukmod, weights=ukmod.person_weight)


def get_test_params(variable):
    test_params = dict(
        ukmod=metadata[variable],
        max_rel_error=MAX_MEAN_REL_ERROR,
        max_mean_rel_error=MAX_MEAN_REL_ERROR,
        min_nonzero_agreement=MIN_NONZERO_AGREEMENT,
        skip_household_matching=False,
    )
    if isinstance(metadata[variable], dict):
        test_params.update(**metadata[variable])
    else:
        test_params["ukmod"] = metadata[variable]
    return test_params


@pytest.mark.parametrize("variable", metadata.keys())
def test_aggregate(variable):
    test_params = get_test_params(variable)
    result = baseline.calc(variable, period=TEST_YEAR).sum()
    target = ukmod[test_params["ukmod"]].sum()

    assert abs(result / target - 1) < test_params["max_rel_error"]

    return result, target


@pytest.mark.parametrize("variable", metadata.keys())
def test_nonzero_count(variable):
    test_params = get_test_params(variable)
    result = (baseline.calc(variable, period=TEST_YEAR) > 0).sum()
    target = (ukmod[test_params["ukmod"]] > 0).sum()

    assert abs(result / target - 1) < test_params["max_rel_error"]

    return result, target


@pytest.mark.parametrize("variable", metadata.keys())
def test_average_error_among_nonzero(variable):
    test_params = get_test_params(variable)
    if test_params["skip_household_matching"]:
        return
    result = pd.Series(
        baseline.calc(variable, period=TEST_YEAR, map_to="household").values,
        index=baseline.calc("household_id", period=TEST_YEAR).values,
    )
    target = ukmod_hh[test_params["ukmod"]]
    # Exclude the smallest 10% to avoid overweighting small errors
    mean_error_ukmod_nonzero = (
        (result / target - 1)[target > target[target > 0].quantile(0.1)]
        .abs()
        .mean()
    )

    assert mean_error_ukmod_nonzero < test_params["max_mean_rel_error"]

    return mean_error_ukmod_nonzero


@pytest.mark.parametrize("variable", metadata.keys())
def test_ukmod_nonzero_agreement(variable):
    test_params = get_test_params(variable)
    result = pd.Series(
        baseline.calc(variable, period=TEST_YEAR, map_to="household").values,
        index=baseline.calc("household_id", period=TEST_YEAR).values,
    )
    target = ukmod_hh[test_params["ukmod"]]
    mean_error = ((result > 0) == (target == 0)).mean()

    assert mean_error < test_params["min_nonzero_agreement"]

    return mean_error



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class student_loans(Variable):
    value_type = float
    entity = Person
    label = u"Student loans"
    definition_period = YEAR


class student_payments(Variable):
    value_type = float
    entity = Person
    label = u"Student payments"
    definition_period = YEAR



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class maternity_allowance_reported(Variable):
    value_type = float
    entity = Person
    label = u"Maternity allowance"
    definition_period = YEAR



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class council_tax_benefit_reported(Variable):
    value_type = float
    entity = Person
    label = u"CTB (reported)"
    definition_period = YEAR


class council_tax_benefit(Variable):
    value_type = float
    entity = BenUnit
    label = u"CTB"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["council_tax_benefit_reported"])



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class JSA_income_reported(Variable):
    value_type = float
    entity = Person
    label = u"JSA (income-based) (reported amount)"
    definition_period = YEAR


class JSA_income_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether eligible for JSA (income-based)"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        JSA = parameters(period).benefit.JSA
        hours = benunit("benunit_weekly_hours", period)
        eligible = benunit("is_single", period) * (
            hours < JSA.hours.single
        ) + benunit("is_couple", period) * (hours < JSA.hours.couple)
        under_SP_age = benunit.min(benunit.members("is_SP_age", period)) == 0
        eligible *= under_SP_age
        employment_statuses = benunit.members("employment_status", period)
        one_unemployed = benunit.any(
            employment_statuses
            == employment_statuses.possible_values.UNEMPLOYED
        )
        eligible *= (
            not_(benunit("income_support", period) > 0) * one_unemployed
        )
        return eligible


class JSA_income_applicable_amount(Variable):
    value_type = float
    entity = BenUnit
    label = u"Maximum amount of JSA (income-based)"
    definition_period = YEAR
    reference = "Jobseekers Act 1995 s. 4"

    def formula(benunit, period, parameters):
        JSA = parameters(period).benefit.JSA
        age = benunit("youngest_adult_age", period)
        personal_allowance = (
            benunit("is_single", period)
            * (
                (age < 25) * JSA.income.amount_18_24
                + (age >= 25) * JSA.income.amount_over_25
            )
            + benunit("is_couple", period) * JSA.income.couple
        ) * WEEKS_IN_YEAR
        premiums = benunit("benefits_premiums", period)
        return (
            (personal_allowance + premiums)
            * benunit("JSA_income_eligible", period)
            * benunit("claims_JSA", period)
        )


class would_claim_JSA(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Would claim income-based JSA"
    documentation = (
        "Whether this family would claim income-based JSA if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return (
            random(benunit) <= parameters(period).benefit.JSA.income.takeup
        ) + benunit("claims_all_entitled_benefits", period)


class claims_JSA(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether this family is imputed to claim JSA based on survey response and take-up rates"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit("would_claim_JSA", period) & benunit(
            "claims_legacy_benefits", period
        )


class JSA_income_applicable_income(Variable):
    value_type = float
    entity = BenUnit
    label = u"Relevant income for JSA (income-based) means test"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        JSA = parameters(period).benefit.JSA
        INCOME_COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "property_income",
            "pension_income",
        ]
        income = aggr(benunit, period, INCOME_COMPONENTS)
        tax = aggr(
            benunit,
            period,
            ["income_tax", "national_insurance"],
        )
        income += aggr(benunit, period, ["social_security_income"])
        income -= tax
        income -= aggr(benunit, period, ["pension_contributions"]) * 0.5
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        income = max_(
            0,
            income
            - (family_type == families.SINGLE)
            * JSA.income.income_disregard_single
            * WEEKS_IN_YEAR
            - benunit("is_couple", period)
            * JSA.income.income_disregard_couple
            * WEEKS_IN_YEAR
            - (family_type == families.LONE_PARENT)
            * JSA.income.income_disregard_lone_parent
            * WEEKS_IN_YEAR,
        )
        return income


class JSA_income(Variable):
    value_type = float
    entity = BenUnit
    label = u"JSA (income-based)"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        amount = benunit("JSA_income_applicable_amount", period)
        income = benunit("JSA_income_applicable_income", period)
        claims_JSA = benunit("claims_JSA", period)
        return (
            max_(
                0,
                (amount - income),
            )
            * claims_JSA
            * benunit("JSA_income_eligible", period)
        )


class JSA(Variable):
    value_type = float
    entity = BenUnit
    label = u"Amount of Jobseeker's Allowance for this family"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["JSA_contrib"]) + benunit(
            "JSA_income", period
        )



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class pension_credit_reported(Variable):
    value_type = float
    entity = Person
    label = u"Reported amount of Pension Credit"
    definition_period = YEAR


class would_claim_PC(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Would claim Pension Credit"
    documentation = (
        "Whether this family would claim Pension Credit if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        takeup_rate = parameters(period).benefit.pension_credit.takeup
        return (random(benunit) < takeup_rate) + benunit(
            "claims_all_entitled_benefits", period
        )


class claims_PC(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether this family is imputed to claim Pension Credit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit("would_claim_PC", period)


class pension_credit_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Eligible for Pension Credit"
    definition_period = YEAR

    def formula(benunit, period):
        all_SP_age = benunit.min(benunit.members("is_SP_age", period))
        one_SP_age = benunit.max(benunit.members("is_SP_age", period))
        claiming_HB = benunit("housing_benefit", period) > 0
        return all_SP_age + (one_SP_age * claiming_HB)


class pension_credit_MG(Variable):
    value_type = float
    entity = BenUnit
    label = u"Pension Credit (Minimum Guarantee) amount per week"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        PC = parameters(period).benefit.pension_credit
        personal_allowance = (
            PC.minimum_guarantee[benunit("relation_type", period)]
            * WEEKS_IN_YEAR
        )
        premiums = benunit("severe_disability_premium", period) + benunit(
            "carer_premium", period
        )
        applicable_amount = personal_allowance + premiums
        return (
            applicable_amount
            * benunit("pension_credit_eligible", period)
            * benunit("claims_PC", period)
        )


class guarantee_credit_applicable_income(Variable):
    value_type = float
    entity = BenUnit
    label = u"Applicable income for Pension Credit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        INCOME_COMPONENTS = [
            "personal_benefits",
            "pension_income",
            "maintenance_income",
            "employment_income",
            "self_employment_income",
            "property_income",
            "savings_interest_income",
            "dividend_income",
        ]
        income = aggr(benunit, period, INCOME_COMPONENTS)
        tax = aggr(
            benunit,
            period,
            ["tax"],
        )
        benefits = add(
            benunit,
            period,
            [
                "child_benefit",
                "child_tax_credit",
                "working_tax_credit",
                "housing_benefit",
            ],
        )
        return amount_over(income + benefits - tax, 0)


class pension_credit_GC(Variable):
    value_type = float
    entity = BenUnit
    label = u"Pension Credit (Guarantee Credit) amount"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        income = benunit("guarantee_credit_applicable_income", period)
        return (
            benunit("pension_credit_eligible", period)
            * benunit("claims_PC", period)
            * max_(0, benunit("pension_credit_MG", period) - income)
        )


class savings_credit_applicable_income(Variable):
    value_type = float
    entity = BenUnit
    label = u"Applicable income for Savings Credit"
    definition_period = YEAR
    reference = "https://www.gov.uk/government/publications/pension-credit-technical-guidance/a-detailed-guide-to-pension-credit-for-advisers-and-others#working-out-income-for-savings-credit"

    def formula(benunit, period, parameters):
        GC_income = benunit("guarantee_credit_applicable_income", period)
        exempted_personal_benefits = aggr(
            benunit,
            period,
            [
                "incapacity_benefit",
                "JSA_contrib",
                "ESA_contrib",
                "SDA",
                "maintenance_income",
            ],
        )
        exempted_family_benefits = add(
            benunit,
            period,
            [
                "working_tax_credit",
            ],
        )
        return amount_over(
            GC_income - exempted_personal_benefits - exempted_family_benefits,
            0,
        )


class pension_credit_SC(Variable):
    value_type = float
    entity = BenUnit
    label = u"Pension Credit (Savings Credit) amount per week"
    definition_period = YEAR
    reference = "https://www.gov.uk/government/publications/pension-credit-technical-guidance/a-detailed-guide-to-pension-credit-for-advisers-and-others#legislation-60-return"

    def formula(benunit, period, parameters):
        PC = parameters(period).benefit.pension_credit
        SC = PC.savings_credit
        income = benunit("savings_credit_applicable_income", period)
        appropriate_amount = benunit("pension_credit_MG", period)
        threshold = (
            SC.income_threshold[benunit("relation_type", period)]
            * WEEKS_IN_YEAR
        )
        maximum_amount = (1 - SC.withdrawal_rate) * max_(
            appropriate_amount - threshold, 0
        )

        amount_A = min_(
            (1 - SC.withdrawal_rate) * max_(income - threshold, 0),
            maximum_amount,
        )
        amount_B = (
            amount_over(income - appropriate_amount, 0) * SC.withdrawal_rate
        )
        amount = max_(amount_A - amount_B, 0)

        return (
            amount
            * benunit("pension_credit_eligible", period)
            * benunit("claims_PC", period)
        )


class pension_credit(Variable):
    value_type = float
    entity = BenUnit
    label = u"Pension Credit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        COMPONENTS = ["pension_credit_GC", "pension_credit_SC"]
        amount = add(benunit, period, COMPONENTS)
        return amount



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class working_tax_credit_reported(Variable):
    value_type = float
    entity = Person
    label = u"Working Tax Credit"
    definition_period = YEAR


class child_tax_credit_reported(Variable):
    value_type = float
    entity = Person
    label = u"Working Tax Credit"
    definition_period = YEAR


class tax_credits_applicable_income(Variable):
    value_type = float
    entity = BenUnit
    label = u"Applicable income for Tax Credits"
    definition_period = YEAR
    reference = "The Tax Credits (Definition and Calculation of Income) Regulations 2002 s. 3"

    def formula(benunit, period, parameters):
        TC = parameters(period).benefit.tax_credits
        STEP_1_COMPONENTS = [
            "pension_income",
            "savings_interest_income",
            "dividend_income",
            "property_income",
        ]
        income = aggr(benunit, period, STEP_1_COMPONENTS)
        income = amount_over(income, TC.means_test.non_earned_disregard)
        STEP_2_COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "social_security_income",
            "miscellaneous_income",
        ]
        income += aggr(benunit, period, STEP_2_COMPONENTS)
        EXEMPT_BENEFITS = ["income_support", "ESA_income", "JSA_income"]
        on_exempt_benefits = add(benunit, period, EXEMPT_BENEFITS) > 0
        return income * not_(on_exempt_benefits)


class is_CTC_child_limit_exempt(Variable):
    value_type = bool
    entity = Person
    label = u"Whether the person was born before 2017 and therefore exempt from the two-child limit for Child Tax Credit"
    definition_period = YEAR

    def formula(person, period, parameters):
        limit_year = parameters(
            period
        ).benefit.tax_credits.child_tax_credit.two_child_limit_year
        born_after = person("birth_year", period.this_year) <= limit_year
        return born_after


class is_child_for_CTC(Variable):
    value_type = bool
    entity = Person
    label = u"Whether this person is a child conferring CTC eligibility"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 8"

    def formula(person, period, parameters):
        return person("is_child_or_QYP", period)


class is_CTC_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether the family is eligible for CTC"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 8"

    def formula(benunit, period, parameters):
        return benunit.sum(benunit.members("is_child_for_CTC", period)) >= 1


class would_claim_CTC(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Would claim Child Tax Credit"
    documentation = (
        "Whether this family would claim Child Tax Credit if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return (
            random(benunit)
            <= parameters(period).benefit.tax_credits.child_tax_credit.takeup
        ) + benunit("claims_all_entitled_benefits", period)


class claims_CTC(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether this family is imputed to claim Child Tax Credit, based on survey response and take-up rates"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit("would_claim_CTC", period) & benunit(
            "claims_legacy_benefits", period
        )


class CTC_maximum_rate(Variable):
    value_type = float
    entity = BenUnit
    label = u"The maximum rate of CTC"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 9"

    def formula(benunit, period, parameters):
        ELEMENTS = [
            "CTC_family_element",
            "CTC_child_element",
            "CTC_disabled_child_element",
            "CTC_severely_disabled_child_element",
        ]
        return add(benunit, period, ELEMENTS)


class CTC_family_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"CTC entitlement in the Family Element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 9"

    def formula(benunit, period, parameters):
        CTC = parameters(period).benefit.tax_credits.child_tax_credit
        return (
            benunit("is_CTC_eligible", period)
            * benunit("claims_CTC", period)
            * CTC.elements.family_element
        )


class CTC_child_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"CTC entitlement from child elements"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 9"

    def formula(benunit, period, parameters):
        exempt_children = benunit.sum(
            benunit.members("is_child_for_CTC", period)
            * benunit.members("is_CTC_child_limit_exempt", period)
        )
        spaces_left = max_(0, 2 - exempt_children)
        non_exempt_children = min_(
            spaces_left,
            benunit.sum(benunit.members("is_child_for_CTC", period)),
        )
        children = exempt_children + non_exempt_children
        CTC = parameters(period).benefit.tax_credits.child_tax_credit
        amount = CTC.elements.child_element * children
        return amount * benunit("claims_CTC", period)


class CTC_disabled_child_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"CTC entitlement from disabled child elements"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 9"

    def formula(benunit, period, parameters):
        disabled_children = benunit.sum(
            benunit.members("is_child_for_CTC", period)
            * (benunit.members("is_disabled_for_benefits", period))
            > 0
        )
        CTC = parameters(period).benefit.tax_credits.child_tax_credit
        amount = CTC.elements.dis_child_element * disabled_children
        return (
            benunit("is_CTC_eligible", period)
            * benunit("claims_CTC", period)
            * amount
        )


class CTC_severely_disabled_child_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"CTC entitlement from severely disabled child elements"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 9"

    def formula(benunit, period, parameters):
        severely_disabled_children = benunit.sum(
            benunit.members("is_child_for_CTC", period)
            * (
                benunit.members("is_severely_disabled_for_benefits", period)
                > 0
            )
        )
        CTC = parameters(period).benefit.tax_credits.child_tax_credit
        amount = (
            CTC.elements.severe_dis_child_element * severely_disabled_children
        )
        return (
            benunit("is_CTC_eligible", period)
            * benunit("claims_CTC", period)
            * amount
        )


class is_WTC_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether the family is eligible for WTC"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 10"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        person_hours = benunit.members("weekly_hours", period)
        hours = benunit.sum(person_hours)
        max_hours = benunit.max(person_hours)
        disabled_adults = benunit("num_disabled_adults", period)
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        has_old = benunit.max(
            benunit.members("age", period.this_year) > WTC.min_hours.old_age
        )
        lower_req = (
            disabled_adults + has_old + (family_type == families.LONE_PARENT)
            > 0
        )
        medium_req = (family_type == families.COUPLE_WITH_CHILDREN) * (
            1 - lower_req
        ) > 0
        higher_req = (benunit("eldest_adult_age", period) >= 25) * (
            benunit("youngest_adult_age", period) < 60
        )
        meets_lower = hours >= WTC.min_hours.lower
        meets_medium = (hours >= WTC.min_hours.couple_with_children) * (
            max_hours >= WTC.min_hours.lower
        )
        meets_higher = hours >= WTC.min_hours.default
        eligible = (
            lower_req * meets_lower
            + medium_req * meets_medium
            + higher_req * meets_higher
        ) > 0
        return eligible


class would_claim_WTC(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Would claim Working Tax Credit"
    documentation = (
        "Whether this family would claim Working Tax Credit if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        takeup_rate = parameters(
            period
        ).benefit.tax_credits.working_tax_credit.takeup
        return benunit("claims_legacy_benefits", period) & (
            random(benunit) < takeup_rate
        ) | benunit("claims_all_entitled_benefits", period)


class claims_WTC(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether this family is imputed to claim Working Tax Credit, based on survey response and take-up rates"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit("would_claim_WTC", period) & benunit(
            "claims_legacy_benefits", period
        )


class WTC_maximum_rate(Variable):
    value_type = float
    entity = BenUnit
    label = u"The maximum rate of WTC"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"

    def formula(benunit, period, parameters):
        ELEMENTS = [
            "WTC_basic_element",
            "WTC_couple_element",
            "WTC_lone_parent_element",
            "WTC_disabled_element",
            "WTC_severely_disabled_element",
            "WTC_worker_element",
            "WTC_childcare_element",
        ]
        return add(benunit, period, ELEMENTS)


class WTC_basic_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"WTC entitlement from the basic element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        return (
            benunit("is_WTC_eligible", period)
            * benunit("claims_WTC", period)
            * WTC.elements.basic
        )


class WTC_couple_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"WTC entitlement from the couple element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        relation_type = benunit("relation_type", period)
        relations = relation_type.possible_values
        amount = (relation_type == relations.COUPLE) * WTC.elements.couple
        return (
            benunit("is_WTC_eligible", period)
            * benunit("claims_WTC", period)
            * amount
        )


class WTC_lone_parent_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"WTC entitlement from the lone parent element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        amount = (
            family_type == families.LONE_PARENT
        ) * WTC.elements.lone_parent
        return (
            benunit("is_WTC_eligible", period)
            * benunit("claims_WTC", period)
            * amount
        )


class WTC_disabled_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"WTC entitlement from the disabled element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        condition = (
            benunit.sum(
                (benunit.members("is_disabled_for_benefits", period) > 0)
                * benunit.members("is_adult", period)
                * benunit.members("weekly_hours", period)
            )
            > WTC.min_hours.lower
        ) > 0
        amount = condition * WTC.elements.disabled
        return (
            benunit("is_WTC_eligible", period)
            * benunit("claims_WTC", period)
            * amount
        )


class WTC_severely_disabled_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"WTC entitlement from the severely disabled element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        amount = (
            benunit("num_severely_disabled_adults", period)
            * WTC.elements.severely_disabled
        )
        return (
            benunit("is_WTC_eligible", period)
            * benunit("claims_WTC", period)
            * amount
        )


class WTC_worker_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"WTC entitlement from the worker element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        hours = benunit.sum(benunit.members("weekly_hours", period))
        amount = (hours > WTC.min_hours.default) * WTC.elements.worker
        return (
            benunit("is_WTC_eligible", period)
            * benunit("claims_WTC", period)
            * amount
        )


class WTC_childcare_element(Variable):
    value_type = float
    entity = BenUnit
    label = u"WTC entitlement from the childcare element"
    definition_period = YEAR
    reference = "Tax Credits Act 2002 s. 11"

    def formula(benunit, period, parameters):
        WTC = parameters(period).benefit.tax_credits.working_tax_credit
        num_children = benunit("num_children", period)
        max_childcare_amount = (
            num_children == 1
        ) * WTC.elements.childcare_1 * WEEKS_IN_YEAR + (
            num_children > 1
        ) * WTC.elements.childcare_2 * WEEKS_IN_YEAR
        childcare_element = min_(
            max_childcare_amount,
            WTC.elements.childcare_coverage
            * benunit.sum(benunit.members("childcare_expenses", period)),
        )
        return (
            benunit("is_WTC_eligible", period)
            * benunit("claims_WTC", period)
            * childcare_element
        )


class tax_credits_reduction(Variable):
    value_type = float
    entity = BenUnit
    label = u"Reduction in Tax Credits from means-tested income"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        means_test = parameters(period).benefit.tax_credits.means_test
        child_tax_credit_amount = benunit("CTC_maximum_rate", period)
        working_tax_credit_amount = benunit("WTC_maximum_rate", period)
        CTC_only = (child_tax_credit_amount > 0) * (
            working_tax_credit_amount == 0
        )
        threshold = select(
            [CTC_only, not_(CTC_only)],
            [
                means_test.income_threshold_CTC_only,
                means_test.income_threshold,
            ],
        )
        income = benunit("tax_credits_applicable_income", period)
        reduction = (
            max_(0, (income - threshold)) * means_test.income_reduction_rate
        )
        return reduction


class working_tax_credit(Variable):
    value_type = float
    entity = BenUnit
    label = u"Working Tax Credit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        amount = max_(
            0,
            benunit("WTC_maximum_rate", period)
            - benunit("tax_credits_reduction", period),
        )
        return amount


class child_tax_credit(Variable):
    value_type = float
    entity = BenUnit
    label = u"Child Tax Credit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        reduction_left = max_(
            0,
            benunit("tax_credits_reduction", period)
            - benunit("WTC_maximum_rate", period),
        )
        amount = max_(
            0,
            benunit("CTC_maximum_rate", period) - reduction_left,
        )
        return amount


class tax_credits(Variable):
    value_type = float
    entity = BenUnit
    label = u"Value of the Tax Credits (benefits) for this family"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("working_tax_credit", period) + person(
            "child_tax_credit", period
        )



from openfisca_uk.model_api import *


class corporate_wealth(Variable):
    label = "Corporate wealth"
    documentation = "Total owned wealth in corporations"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"


class corporate_tax_incidence(Variable):
    label = "Corporate tax incidence"
    documentation = (
        "Reduction in value of corporate wealth due to taxes on corporations"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period):
        total_change = household.state(
            "corporate_stamp_duty_revenue_change", period
        )
        total_wealth = (
            household("corporate_wealth", period)
            * household("household_weight", period)
        ).sum()
        percentage_change = where(
            total_wealth > 0, total_change / total_wealth, 0
        )
        return percentage_change * household("corporate_wealth", period)



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class is_disabled_for_benefits(Variable):
    value_type = bool
    entity = Person
    label = u"Has a disability"
    documentation = "Whether this person is disabled for benefits purposes"
    definition_period = YEAR
    reference = "Child Tax Credit Regulations 2002 s. 8"

    def formula(person, period, parameters):
        QUALIFYING_BENEFITS = [
            "DLA_M",
            "DLA_SC",
            "PIP_M",
            "PIP_DL",
        ]
        return add(person, period, QUALIFYING_BENEFITS) > 0


class is_enhanced_disabled_for_benefits(Variable):
    value_type = bool
    entity = Person
    label = u"Whether meets the middle disability benefit entitlement"
    definition_period = YEAR

    def formula(person, period, parameters):
        DLA_requirement = (
            parameters(period).benefit.DLA.self_care.highest * WEEKS_IN_YEAR
        )
        sufficient_DLA = person("DLA_SC", period) >= DLA_requirement
        return sufficient_DLA


class is_severely_disabled_for_benefits(Variable):
    value_type = bool
    entity = Person
    label = u"Has a severe disability"
    documentation = (
        "Whether this person is severely disabled for benefits purposes"
    )
    definition_period = YEAR
    reference = "Child Tax Credit Regulations 2002 s. 8"

    def formula(person, period, parameters):
        benefit = parameters(period).benefit
        THRESHOLD_SAFETY_GAP = 10 * WEEKS_IN_YEAR
        paragraph_3 = (
            person("DLA_SC", period)
            >= benefit.DLA.self_care.highest * WEEKS_IN_YEAR
            - THRESHOLD_SAFETY_GAP
        )
        paragraph_4 = (
            person("PIP_DL", period)
            >= benefit.PIP.daily_living.higher * WEEKS_IN_YEAR
            - THRESHOLD_SAFETY_GAP
        )
        paragraph_5 = person("AFCS", period) > 0
        return sum([paragraph_3, paragraph_4, paragraph_5]) > 0


class num_disabled_children(Variable):
    value_type = int
    entity = BenUnit
    label = u"Number of disabled children"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        child = benunit.members("is_child_or_QYP", period)
        disabled = benunit.members("is_disabled_for_benefits", period) > 0
        return benunit.sum(child * disabled)


class num_enhanced_disabled_children(Variable):
    value_type = int
    entity = BenUnit
    label = u"Number of enhanced disabled children"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        child = benunit.members("is_child_or_QYP", period)
        disabled = (
            benunit.members("is_enhanced_disabled_for_benefits", period) > 0
        )
        return benunit.sum(child * disabled)


class num_severely_disabled_children(Variable):
    value_type = int
    entity = BenUnit
    label = u"Number of severely disabled children"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        child = benunit.members("is_child_or_QYP", period)
        disabled = (
            benunit.members("is_severely_disabled_for_benefits", period) > 0
        )
        return benunit.sum(child * disabled)


class num_disabled_adults(Variable):
    value_type = int
    entity = BenUnit
    label = u"Number of disabled adults"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        adult = benunit.members("is_adult", period)
        disabled = benunit.members("is_disabled_for_benefits", period) > 0
        return benunit.sum(adult * disabled)


class num_enhanced_disabled_adults(Variable):
    value_type = int
    entity = BenUnit
    label = u"Number of enhanced disabled adults"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        adult = benunit.members("is_adult", period)
        disabled = (
            benunit.members("is_enhanced_disabled_for_benefits", period) > 0
        )
        return benunit.sum(adult * disabled)


class num_severely_disabled_adults(Variable):
    value_type = int
    entity = BenUnit
    label = u"Number of severely disabled adults"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        adult = benunit.members("is_adult", period)
        disabled = (
            benunit.members("is_severely_disabled_for_benefits", period) > 0
        )
        return benunit.sum(adult * disabled)


class disability_premium(Variable):
    value_type = float
    entity = BenUnit
    label = u"Disability premium"
    definition_period = YEAR
    reference = "The Social Security Amendment (Enhanced Disability Premium) Regulations 2000"

    def formula(benunit, period, parameters):
        dis = parameters(period).benefit.disability_premia
        amount = (
            (benunit("num_disabled_adults", period.this_year) > 0)
            * (
                benunit("is_single", period.this_year) * dis.disability_single
                + benunit("is_couple", period.this_year)
                * dis.disability_couple
            )
            * WEEKS_IN_YEAR
        )
        return amount


class severe_disability_premium(Variable):
    value_type = float
    entity = BenUnit
    label = u"Severe disability premium"
    definition_period = YEAR
    reference = "The Social Security Amendment (Enhanced Disability Premium) Regulations 2000"

    def formula(benunit, period, parameters):
        dis = parameters(period).benefit.disability_premia
        amount = (
            (benunit("num_severely_disabled_adults", period.this_year) > 0)
            * (
                benunit("is_single", period.this_year) * dis.severe_single
                + benunit("is_couple", period.this_year) * dis.severe_couple
            )
            * WEEKS_IN_YEAR
        )
        return amount


class enhanced_disability_premium(Variable):
    value_type = float
    entity = BenUnit
    label = u"Enhanced disability premium"
    definition_period = YEAR
    reference = "The Social Security Amendment (Enhanced Disability Premium) Regulations 2000"

    def formula(benunit, period, parameters):
        dis = parameters(period).benefit.disability_premia
        amount = (
            (benunit("num_enhanced_disabled_adults", period.this_year) > 0)
            * (
                benunit("is_single", period.this_year) * dis.enhanced_single
                + benunit("is_couple", period.this_year) * dis.enhanced_couple
            )
            * WEEKS_IN_YEAR
        )
        return amount



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class JSA_contrib(Variable):
    value_type = float
    entity = Person
    label = u"JSA (contribution-based)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("JSA_contrib_reported", period)


class JSA_contrib_reported(Variable):
    value_type = float
    entity = Person
    label = u"Job Seeker's Allowance (contribution-based) (reported)"
    definition_period = YEAR



from openfisca_uk.model_api import *


class owned_land(Variable):
    entity = Household
    label = "Owned land"
    documentation = "Total value of all land-only plots owned by the household"
    unit = "currency-GBP"
    definition_period = YEAR
    value_type = float


class land_value(Variable):
    entity = Household
    label = "Land value"
    documentation = "Estimated total land value exposure (your property's land value, and any share of corporate land value)"
    unit = "currency-GBP"
    definition_period = YEAR
    value_type = float

    def formula(household, period, parameters):
        property_wealth = household("property_wealth", period)
        corporate_wealth = household("corporate_wealth", period)
        total_property_wealth = (
            property_wealth * household("household_weight", period)
        ).sum()
        total_corporate_wealth = (
            corporate_wealth * household("household_weight", period)
        ).sum()
        land = parameters(period).wealth.land
        property_wealth_intensity = (
            land.value.aggregate_household_land_value / total_property_wealth
        )
        property_wealth_intensity = where(
            total_property_wealth > 0,
            property_wealth_intensity,
            land.intensity.property_wealth,
        )
        corporate_wealth_intensity = (
            land.value.aggregate_corporate_land_value / total_corporate_wealth
        )
        corporate_wealth_intensity = where(
            total_corporate_wealth > 0,
            corporate_wealth_intensity,
            land.intensity.corporate_wealth,
        )
        return (
            property_wealth * property_wealth_intensity
            + corporate_wealth * corporate_wealth_intensity
            + household("owned_land", period)
        )



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class household_id(Variable):
    value_type = int
    entity = Household
    label = u"ID for the household"
    definition_period = YEAR


class households(Variable):
    value_type = float
    entity = Household
    label = u"Variable holding households"
    definition_period = YEAR
    default_value = 1


class num_bedrooms(Variable):
    value_type = int
    entity = Household
    label = u"The number of bedrooms in the house"
    definition_period = YEAR


class is_shared_accommodation(Variable):
    value_type = bool
    entity = Household
    label = u"Whether the household is shared accommodation"
    definition_period = YEAR


class household_weight(Variable):
    value_type = float
    entity = Household
    label = u"Weight factor for the household"
    definition_period = YEAR


class Country(Enum):
    UNKNOWN = u"Unknown"
    SCOTLAND = u"Scotland"
    ENGLAND = u"England"
    WALES = u"Wales"
    NORTHERN_IRELAND = u"Northern Ireland"


class country(Variable):
    value_type = Enum
    possible_values = Country
    default_value = Country.ENGLAND
    entity = Household
    label = u"Country of the UK"
    definition_period = ETERNITY

    def formula(household, period, parameters):
        unknown_region = household("region", period) == Region.UNKNOWN
        wales = household("region", period) == Region.WALES
        ni = household("region", period) == Region.NORTHERN_IRELAND
        scot = household("region", period) == Region.SCOTLAND
        country = select(
            [unknown_region, scot, wales, ni, True],
            [
                Country.UNKNOWN,
                Country.SCOTLAND,
                Country.WALES,
                Country.NORTHERN_IRELAND,
                Country.ENGLAND,
            ],
        )
        return country


class Region(Enum):
    UNKNOWN = u"Unknown"
    NORTH_EAST = u"North East"
    NORTH_WEST = u"North West"
    YORKSHIRE = u"Yorkshire and the Humber"
    EAST_MIDLANDS = u"East Midlands"
    WEST_MIDLANDS = u"West Midlands"
    EAST_OF_ENGLAND = u"East of England"
    LONDON = u"London"
    SOUTH_EAST = u"South East"
    SOUTH_WEST = u"South West"
    WALES = u"Wales"
    SCOTLAND = u"Scotland"
    NORTHERN_IRELAND = u"Northern Ireland"


class region(Variable):
    value_type = Enum
    possible_values = Region
    default_value = Region.LONDON
    entity = Household
    label = u"Region"
    documentation = "Area of the UK"
    definition_period = ETERNITY


class TenureType(Enum):
    RENT_FROM_COUNCIL = "Rented from Council"
    RENT_FROM_HA = "Rented from a Housing Association"
    RENT_PRIVATELY = "Rented privately"
    OWNED_OUTRIGHT = "Owned outright"
    OWNED_WITH_MORTGAGE = "Owned with a mortgage"


class tenure_type(Variable):
    value_type = Enum
    possible_values = TenureType
    default_value = TenureType.RENT_PRIVATELY
    entity = Household
    label = u"Tenure type of the household"
    definition_period = YEAR


class is_renting(Variable):
    value_type = bool
    entity = Household
    label = u"Is renting"
    definition_period = YEAR

    def formula(household, period, parameters):
        tenure = household("tenure_type", period)
        rent_types = (
            TenureType.RENT_PRIVATELY,
            TenureType.RENT_FROM_COUNCIL,
            TenureType.RENT_PRIVATELY,
        )
        return sum([tenure == rent_type for rent_type in rent_types])


class AccommodationType(Enum):
    HOUSE_DETACHED = "Detached house"
    HOUSE_SEMI_DETACHED = "Semi-detached house"
    HOUSE_TERRACED = "Terraced house"
    FLAT = "Purpose-built flat or maisonette"
    CONVERTED_HOUSE = "Converted house or building"
    MOBILE = "Caravan/Mobile home or houseboat"
    OTHER = "Other"
    UNKNOWN = "Unknown"


class accommodation_type(Variable):
    value_type = Enum
    possible_values = AccommodationType
    default_value = AccommodationType.UNKNOWN
    entity = Household
    label = u"Type of accommodation"
    definition_period = ETERNITY


class household_equivalisation_bhc(Variable):
    value_type = float
    entity = Household
    label = u"Equivalisation factor to account for household composition, before housing costs"
    definition_period = YEAR

    def formula(household, period, parameters):
        other_adults = max_(
            household.sum(household.members("is_adult", period)) - 1, 0
        )
        num_young_children = household.sum(
            household.members("is_young_child", period)
        )
        num_older_children = household.sum(
            household.members("is_older_child", period)
        )
        weighting = (
            0.67
            + 0.33 * other_adults
            + 0.33 * num_older_children
            + 0.2 * num_young_children
        )
        return weighting


class household_equivalisation_ahc(Variable):
    value_type = float
    entity = Household
    label = u"Equivalisation factor to account for household composition, after housing costs"
    definition_period = YEAR

    def formula(household, period, parameters):
        other_adults = max_(
            household.sum(household.members("is_adult", period)) - 1, 0
        )
        num_young_children = household.sum(
            household.members("is_young_child", period)
        )
        num_older_children = household.sum(
            household.members("is_older_child", period)
        )
        weighting = (
            0.58
            + 0.42 * other_adults
            + 0.42 * num_older_children
            + 0.2 * num_young_children
        )
        return weighting


class household_num_people(Variable):
    value_type = int
    entity = Household
    label = u"Number of people"
    definition_period = YEAR
    unit = "person"

    def formula(household, period, parameters):
        return household.nb_persons()


class household_num_benunits(Variable):
    value_type = int
    entity = Household
    label = u"Number of benefit units"
    definition_period = YEAR
    unit = "benefit unit"

    def formula(household, period, parameters):
        return household.sum(household.members("is_benunit_head", period))


class household_random_number(Variable):
    label = "Randomness"
    documentation = (
        "A random number between zero and one generated for this household"
    )
    entity = Household
    definition_period = YEAR
    value_type = float

    def formula(household, period):
        return random(household)



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *

"""
This file contains variables that are commonly used in benefit eligibility calculations.
"""


class family_benefits(Variable):
    value_type = float
    entity = Person
    label = u"Total simulated family benefits for this person"
    definition_period = YEAR

    def formula(person, period, parameters):
        FAMILY_BENEFITS = [
            "child_benefit",
            "ESA_income",
            "housing_benefit",
            "income_support",
            "JSA_income",
            "pension_credit",
            "universal_credit",
            "council_tax_benefit",
        ]
        benefits = add(person.benunit, period, FAMILY_BENEFITS) * person(
            "is_benunit_head", period
        )
        benefits += add(
            person.benunit, period, ["working_tax_credit", "child_tax_credit"]
        ) * person("is_benunit_head", period)
        return benefits


class family_benefits_reported(Variable):
    value_type = float
    entity = Person
    label = u"Total reported family benefits for this person"
    definition_period = YEAR

    def formula(person, period, parameters):
        FAMILY_BENEFITS = [
            "child_benefit",
            "ESA_income",
            "housing_benefit",
            "income_support",
            "JSA_income",
            "pension_credit",
            "universal_credit",
        ]
        benefits = add(
            person,
            period,
            map(lambda ben: ben + "_reported", FAMILY_BENEFITS),
        )
        benefits += add(
            person,
            period,
            ["working_tax_credit_reported", "child_tax_credit_reported"],
        )
        return benefits


class benefits(Variable):
    value_type = float
    entity = Person
    label = u"Total benefits"
    documentation = "Total state benefits"
    unit = "currency-GBP"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("personal_benefits", period) + person(
            "family_benefits", period
        )


class household_benefits(Variable):
    value_type = float
    entity = Household
    label = u"Benefits"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(household, period, parameters):
        return household.sum(household.members("benefits", period))


class other_benefits(Variable):
    value_type = float
    entity = Person
    label = u"Income from benefits not modelled or detailed in the model"
    definition_period = YEAR

    def formula(person, period, parameters):
        return (
            person("benefits_reported", period)
            - person("personal_benefits", period)
            - person("family_benefits", period)
        )


class benefits_reported(Variable):
    value_type = float
    entity = Person
    label = u"Total simulated"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("personal_benefits_reported", period) + person(
            "family_benefits_reported", period
        )


class benefits_modelling(Variable):
    value_type = float
    entity = Person
    label = (
        u"Difference between reported and simulated benefits for this person"
    )
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("benefits", period) - person("benefits_reported", period)


class is_QYP(Variable):
    value_type = bool
    entity = Person
    label = u"Whether this person is a qualifying young person for benefits purposes"
    definition_period = YEAR

    def formula(person, period, parameters):
        education = person("current_education", period)
        return (person("age", period) < 20) & ~(
            education == education.possible_values.NOT_IN_EDUCATION
        )


class is_child_or_QYP(Variable):
    value_type = bool
    entity = Person
    label = u"Whether this person is a child or qualifying young person for most benefits"
    definition_period = YEAR

    def formula(person, period, parameters):
        return (person("age", period) < 16) | person("is_QYP", period)


class benefits_premiums(Variable):
    value_type = float
    entity = BenUnit
    label = u"Value of premiums for disability and carer status"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        PREMIUMS = [
            "disability_premium",
            "enhanced_disability_premium",
            "severe_disability_premium",
            "carer_premium",
        ]
        return add(benunit, period, PREMIUMS)


class benunit_weekly_hours(Variable):
    value_type = float
    entity = BenUnit
    label = u"Average weekly hours worked by adults in the benefit unit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.sum(benunit.members("weekly_hours", period))


class is_single(Variable):
    value_type = bool
    entity = BenUnit
    label = (
        u"Whether this benefit unit contains a single claimant for benefits"
    )
    definition_period = ETERNITY

    def formula(benunit, period, parameters):
        relation_type = benunit("relation_type", period)
        relations = relation_type.possible_values
        return relation_type == relations.SINGLE


class is_couple(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether this benefit unit contains a joint couple claimant for benefits"
    definition_period = ETERNITY

    def formula(benunit, period, parameters):
        relation_type = benunit("relation_type", period)
        relations = relation_type.possible_values
        return relation_type == relations.COUPLE


class is_lone_parent(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether the family is a lone parent family"
    definition_period = ETERNITY

    def formula(benunit, period, parameters):
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        return family_type == families.LONE_PARENT


class is_single_person(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Whether the family is a single person"
    definition_period = ETERNITY

    def formula(benunit, period, parameters):
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        return family_type == families.SINGLE


class personal_benefits(Variable):
    value_type = float
    entity = Person
    label = u"Value of personal, non-means-tested benefits"
    definition_period = YEAR

    def formula(person, period, parameters):
        BENEFITS = [
            "AA",
            "AFCS",
            "BSP",
            "carers_allowance",
            "DLA_M",
            "DLA_SC",
            "ESA_contrib",
            "IIDB",
            "incapacity_benefit",
            "JSA_contrib",
            "PIP_M",
            "PIP_DL",
            "SDA",
            "state_pension",
        ]
        return add(person, period, BENEFITS)


class personal_benefits_reported(Variable):
    value_type = float
    entity = Person
    label = u"Value of personal, non-means-tested benefits"
    definition_period = YEAR

    def formula(person, period, parameters):
        BENEFITS = [
            "AA",
            "AFCS",
            "BSP",
            "carers_allowance",
            "DLA_M",
            "DLA_SC",
            "ESA_contrib",
            "IIDB",
            "incapacity_benefit",
            "JSA_contrib",
            "PIP_M",
            "PIP_DL",
            "SDA",
            "state_pension",
        ]
        return add(
            person, period, map(lambda ben: ben + "_reported", BENEFITS)
        )


class claims_all_entitled_benefits(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Claims all eligible benefits"
    definition_period = YEAR
    documentation = (
        "Whether this family would claim any benefit they are entitled to"
    )


class claims_legacy_benefits(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Claims legacy benefits"
    documentation = "Whether this family is currently receiving legacy benefits (overrides UC claimant status)"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        # assign legacy/UC claimant status, consistently for each household
        return (
            benunit.value_from_first_person(
                benunit.members.household.project(
                    random(benunit.members.household)
                )
            )
            > parameters(period).benefit.universal_credit.rollout_rate
        )



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class AFCS(Variable):
    value_type = float
    entity = Person
    label = u"Armed Forces Compensation Scheme"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("AA_reported", period)


class AFCS_reported(Variable):
    value_type = float
    entity = Person
    label = u"Armed Forces Compensation Scheme (reported)"
    definition_period = YEAR



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class earned_income(Variable):
    value_type = float
    entity = Person
    label = u"Total earned income"
    definition_period = YEAR

    def formula(person, period, parameters):
        COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "pension_income",
        ]
        return add(person, period, COMPONENTS)


class sublet_income(Variable):
    value_type = float
    entity = Person
    label = u"Income received from sublet agreements"
    definition_period = YEAR


class miscellaneous_income(Variable):
    value_type = float
    entity = Person
    label = u"Income from other sources"
    definition_period = YEAR


class private_transfer_income(Variable):
    value_type = float
    entity = Person
    label = u"Private transfers"
    definition_period = YEAR


class lump_sum_income(Variable):
    value_type = float
    entity = Person
    label = u"Lump sum income"
    definition_period = YEAR


class market_income(Variable):
    value_type = float
    entity = Person
    label = u"Market income"
    definition_period = YEAR

    def formula(person, period, parameters):
        return (
            add(
                person,
                period,
                [
                    "employment_income",
                    "self_employment_income",
                    "savings_interest_income",
                    "dividend_income",
                    "miscellaneous_income",
                    "property_income",
                    "pension_income",
                    "private_transfer_income",
                    "maintenance_income",
                ],
            )
            - person("maintenance_expenses", period)
        )


class gross_income(Variable):
    value_type = float
    entity = Person
    label = u"Gross income, including benefits"
    definition_period = YEAR

    def formula(person, period, parameters):
        COMPONENTS = [
            "employment_income",
            "pension_income",
            "self_employment_income",
            "property_income",
            "savings_interest_income",
            "dividend_income",
            "miscellaneous_income",
            "benefits",
        ]
        return add(person, period, COMPONENTS)


class household_gross_income(Variable):
    value_type = float
    entity = Household
    unit = "currency-GBP"
    label = u"Household gross income"
    definition_period = YEAR

    def formula(household, period, parameters):
        return household.sum(household.members("gross_income", period))


class net_income(Variable):
    value_type = float
    entity = Person
    label = u"Net income"
    documentation = "Market income, minus taxes, plus benefits"
    unit = "currency-GBP"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("gross_income", period) - person("tax", period)


class hours_worked(Variable):
    value_type = float
    entity = Person
    label = u"Total amount of hours worked by this person"
    definition_period = YEAR


class in_work(Variable):
    value_type = bool
    entity = Person
    label = u"Worked some hours"
    definition_period = YEAR

    def formula(person, period, parameters):
        has_hours_worked = person("hours_worked", period) > 0
        has_earnings = (
            add(
                person, period, ["employment_income", "self_employment_income"]
            )
            > 0
        )
        return has_hours_worked | has_earnings


class weekly_hours(Variable):
    value_type = float
    entity = Person
    label = u"Weekly hours"
    documentation = "Average weekly hours worked"
    definition_period = YEAR
    unit = "hour"

    def formula(person, period, parameters):
        return person("hours_worked", period) / WEEKS_IN_YEAR


class EmploymentStatus(Enum):
    FT_EMPLOYED = "Full-time employed"
    PT_EMPLOYED = "Part-time employed"
    FT_SELF_EMPLOYED = "Full-time self-employed"
    PT_SELF_EMPLOYED = "Part-time self-employed"
    UNEMPLOYED = "Unemployed"
    RETIRED = "Retired"
    STUDENT = "Student"
    CARER = "Carer"
    LONG_TERM_DISABLED = "Long-term sick/disabled"
    SHORT_TERM_DISABLED = "Short-term sick/disabled"
    OTHER_INACTIVE = "Inactive for another reason"
    CHILD = "Child"


class employment_status(Variable):
    value_type = Enum
    entity = Person
    possible_values = EmploymentStatus
    default_value = EmploymentStatus.UNEMPLOYED
    label = u"Employment status of the person"
    definition_period = YEAR


class capital_income(Variable):
    value_type = float
    entity = Person
    label = u"Income from savings or dividends"
    definition_period = YEAR

    def formula(person, period, parameters):
        return add(
            person, period, ["savings_interest_income", "dividend_income"]
        )


class maintenance_income(Variable):
    value_type = float
    entity = Person
    label = u"Maintenance payments"
    definition_period = YEAR


class hbai_household_net_income(Variable):
    value_type = float
    entity = Household
    label = u"Household net income (HBAI definition)"
    documentation = "Disposable income for the household, following the definition used for official poverty statistics"
    unit = "currency-GBP"
    definition_period = YEAR

    def formula(household, period, parameters):
        gross_income = household("household_gross_income", period)
        tax = household("household_tax", period)
        ignored_taxes = household("expected_stamp_duty", period) + household(
            "corporate_tax_incidence", period
        )
        return gross_income - tax + ignored_taxes


class household_net_income(Variable):
    label = "Household net income"
    documentation = "Disposable income for the household"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period):
        gross_income = household("household_gross_income", period)
        tax = household("household_tax", period)
        return gross_income - tax


class hbai_household_net_income_ahc(Variable):
    value_type = float
    entity = Household
    label = u"Household net income, after housing costs"
    definition_period = YEAR

    def formula(household, period, parameters):
        return household("hbai_household_net_income", period) - household(
            "housing_costs", period
        )


class equiv_household_net_income(Variable):
    value_type = float
    entity = Household
    label = u"Equivalised household net income"
    definition_period = YEAR

    def formula(household, period, parameters):
        return household("household_net_income", period) / household(
            "household_equivalisation_bhc", period
        )


class equiv_hbai_household_net_income(Variable):
    value_type = float
    entity = Household
    label = u"Equivalised household net income (HBAI)"
    definition_period = YEAR

    def formula(household, period, parameters):
        return household("hbai_household_net_income", period) / household(
            "household_equivalisation_bhc", period
        )


class equiv_hbai_household_net_income_ahc(Variable):
    value_type = float
    entity = Household
    label = u"Equivalised household net income, after housing costs (HBAI)"
    definition_period = YEAR

    def formula(household, period, parameters):
        return household("hbai_household_net_income_ahc", period) / household(
            "household_equivalisation_ahc", period
        )


class base_net_income(Variable):
    value_type = float
    entity = Person
    label = "Existing net income for the person to use as a base in microsimulation"
    definition_period = YEAR


class is_apprentice(Variable):
    value_type = bool
    entity = Person
    label = u"In an apprenticeship programme"
    definition_period = YEAR
    default_value = False


class MinimumWageCategory(Enum):
    APPRENTICE = "Apprentice"
    UNDER_18 = "Under 18"
    BETWEEN_18_20 = "18 to 20"
    BETWEEN_21_22 = "21 to 22"
    BETWEEN_23_24 = "23 to 24"
    OVER_24 = "25 or over"


class minimum_wage_category(Variable):
    value_type = Enum
    possible_values = MinimumWageCategory
    default_value = MinimumWageCategory.OVER_24
    entity = Person
    label = "Minimum wage category"
    definition_period = YEAR

    def formula(person, period, parameters):
        age = person("age", period)
        return select(
            [
                person("is_apprentice", period),
                age < 18,
                (age >= 18) & (age <= 20),
                (age >= 21) & (age <= 22),
                (age >= 23) & (age <= 24),
                age >= 25,
            ],
            [
                MinimumWageCategory.APPRENTICE,
                MinimumWageCategory.UNDER_18,
                MinimumWageCategory.BETWEEN_18_20,
                MinimumWageCategory.BETWEEN_21_22,
                MinimumWageCategory.BETWEEN_23_24,
                MinimumWageCategory.OVER_24,
            ],
        )


class minimum_wage(Variable):
    value_type = float
    entity = Person
    label = u"Minimum wage"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(person, period, parameters):
        MW = parameters(period).law.minimum_wage
        return MW[person("minimum_wage_category", period)]


class household_market_income(Variable):
    value_type = float
    entity = Household
    label = u"Household market income"
    documentation = "Market income for the household"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(household, period, parameters):
        return household.sum(household.members("market_income", period))



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class tax(Variable):
    value_type = float
    entity = Person
    label = u"Total tax"
    definition_period = YEAR

    def formula(person, period, parameters):
        return add(
            person,
            period,
            ["income_tax", "national_insurance"],
        )


class household_tax(Variable):
    value_type = float
    entity = Household
    label = u"Total tax"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(household, period):
        personal_taxes = household.sum(household.members("tax", period))
        household_taxes = (
            household("expected_stamp_duty", period)
            + household("corporate_tax_incidence", period)
            + household("council_tax", period)
        )
        return personal_taxes + household_taxes


class benunit_tax(Variable):
    value_type = float
    entity = BenUnit
    label = u"Benefit unit tax paid"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.sum(benunit.members("tax", period))


class tax_reported(Variable):
    value_type = float
    entity = Person
    label = u"Reported tax paid"
    definition_period = YEAR


class tax_modelling(Variable):
    value_type = float
    entity = Person
    label = u"Difference between reported and imputed tax liabilities"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("tax", period) - person("tax_reported", period)



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class ESA_income_reported(Variable):
    value_type = float
    entity = Person
    label = u"ESA (income-based) (reported amount)"
    definition_period = YEAR


class would_claim_ESA_income(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Would claim income-based ESA"
    documentation = "Whether this family would claim income-based Employment Support Allowance if eligible"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return (
            aggr(benunit, period, ["ESA_income_reported"])
            + benunit("claims_all_entitled_benefits", period)
            > 0
        )


class ESA_income_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = u"ESA (income) eligible"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["ESA_income_reported"]) > 0


class claims_ESA_income(Variable):
    value_type = bool
    entity = BenUnit
    label = u"Claims ESA (income)"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit("would_claim_ESA_income", period) & benunit(
            "claims_legacy_benefits", period
        )


class ESA_income(Variable):
    value_type = float
    entity = BenUnit
    label = u"ESA (income-based)"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["ESA_income_reported"])



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class tax(Variable):
    value_type = float
    entity = Person
    label = u"Total tax liability"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("tax_reported", period)


class reported_tax(Reform):
    def apply(self):
        self.update_variable(tax)


class child_benefit(Variable):
    value_type = float
    entity = BenUnit
    label = u"Child Benefit entitlement for the family"
    definition_period = YEAR
    reference = "Social Security Contributions and Benefits Act 1992 s. 141"

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["child_benefit_reported"])


class ESA_income(Variable):
    value_type = float
    entity = BenUnit
    label = u"ESA (income-based)"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["ESA_income_reported"])


class housing_benefit(Variable):
    value_type = float
    entity = BenUnit
    label = u"Housing Benefit for the family"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["housing_benefit_reported"])


class income_support(Variable):
    value_type = float
    entity = BenUnit
    label = u"Income Support"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["income_support_reported"])


class JSA_income(Variable):
    value_type = float
    entity = BenUnit
    label = u"Job Seeker's Allowance (income-based)"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["JSA_income_reported"])


class pension_credit(Variable):
    value_type = float
    entity = BenUnit
    label = u"Pension credit amount"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["pension_credit_reported"])


class working_tax_credit(Variable):
    value_type = float
    entity = BenUnit
    label = u"Working Tax Credit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["working_tax_credit_reported"])


class child_tax_credit(Variable):
    value_type = float
    entity = BenUnit
    label = u"Amount of Child Tax Credit entitled to"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["child_tax_credit_reported"])


class universal_credit(Variable):
    value_type = float
    entity = BenUnit
    label = u"Universal Credit amount"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["universal_credit_reported"])


class benefits(Variable):
    value_type = float
    entity = Person
    label = u"Total simulated"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("benefits_reported", period)


class AA(Variable):
    value_type = float
    entity = Person
    label = u"Attendance Allowance"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("AA_reported", period)


class AFCS(Variable):
    value_type = float
    entity = Person
    label = u"Armed Forces Compensation Scheme"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("AA_reported", period)


class BSP(Variable):
    value_type = float
    entity = Person
    label = u"Bereavement Support Payment"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("BSP_reported", period)


class carers_allowance(Variable):
    value_type = float
    entity = Person
    label = u"Carer's Allowance"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("carers_allowance_reported", period)


class DLA_M(Variable):
    value_type = float
    entity = Person
    label = u"Disability Living Allowance (mobility component)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("DLA_M_reported", period)


class DLA_SC(Variable):
    value_type = float
    entity = Person
    label = u"Disability Living Allowance (self-care)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("DLA_SC_reported", period)


class ESA_contrib(Variable):
    value_type = float
    entity = Person
    label = u"Employment and Support Allowance (contribution-based)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("ESA_contrib_reported", period)


class IIDB(Variable):
    value_type = float
    entity = Person
    label = u"Industrial Injuries Disablement Benefit"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("IIDB_reported", period)


class incapacity_benefit(Variable):
    value_type = float
    entity = Person
    label = u"Incapacity Benefit"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("incapacity_benefit_reported", period)


class JSA_contrib(Variable):
    value_type = float
    entity = Person
    label = u"Job Seeker's Allowance (contribution-based)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("JSA_contrib_reported", period)


class PIP_DL(Variable):
    value_type = float
    entity = Person
    label = u"Personal Independence Payment (Daily Living)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("PIP_DL_reported", period)


class PIP_M(Variable):
    value_type = float
    entity = Person
    label = u"Personal Independence Payment (Mobility)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("PIP_M_reported", period)


class SDA(Variable):
    value_type = float
    entity = Person
    label = u"Severe Disablement Allowance"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("SDA_reported", period)


class state_pension(Variable):
    value_type = float
    entity = Person
    label = u"Income from the State Pension"
    definition_period = YEAR
    unit = "currency-GBP"
    documentation = "Gross State Pension payments"

    def formula(person, period, parameters):
        return person("state_pension_reported", period)


class reported_benefits(Reform):
    def apply(self):
        for var in [
            child_benefit,
            ESA_income,
            housing_benefit,
            income_support,
            JSA_income,
            pension_credit,
            working_tax_credit,
            child_tax_credit,
            universal_credit,
            AA,
            AFCS,
            BSP,
            carers_allowance,
            DLA_M,
            DLA_SC,
            ESA_contrib,
            IIDB,
            incapacity_benefit,
            JSA_contrib,
            PIP_M,
            PIP_DL,
            SDA,
            state_pension,
            benefits,
        ]:
            self.update_variable(var)


reported = (reported_tax, reported_benefits)



from openfisca_uk import CountryTaxBenefitSystem
from openfisca_uk.entities import entities
import numpy as np
import warnings
from openfisca_uk.entities import *
import numpy as np
import warnings
from openfisca_uk.tools.parameters import backdate_parameters
from openfisca_uk_data import DATASETS
from openfisca_tools.microsimulation import (
    Microsimulation as GeneralMicrosimulation,
)
from openfisca_tools.hypothetical import IndividualSim as GeneralIndividualSim
import yaml
from pathlib import Path


with open(Path(__file__).parent / "datasets.yml") as f:
    datasets = yaml.safe_load(f)
    DEFAULT_DATASET = list(
        filter(lambda ds: ds.name == datasets["default"], DATASETS)
    )[0]


warnings.filterwarnings("ignore")

np.random.seed(0)


class Microsimulation(GeneralMicrosimulation):
    tax_benefit_system = CountryTaxBenefitSystem
    entities = entities
    default_dataset = DEFAULT_DATASET
    post_reform = backdate_parameters()


class IndividualSim(GeneralIndividualSim):
    tax_benefit_system = CountryTaxBenefitSystem
    post_reform = backdate_parameters()



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class weekly_rent(Variable):
    value_type = float
    entity = Household
    label = u"Weekly average rent"
    definition_period = YEAR


@uprated(by="CPI")
class benunit_rent(Variable):
    value_type = float
    entity = BenUnit
    label = u"Rent"
    documentation = "Gross rent that members of this family are liable for"
    definition_period = YEAR
    unit = "currency-GBP"

    def formula(benunit, period, parameters):
        return benunit.sum(benunit.members("personal_rent", period))


class personal_rent(Variable):
    value_type = float
    entity = Person
    label = u"Rent liable"
    documentation = "The gross rent this person is liable for"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person.household("rent", period) * person(
            "is_household_head", period
        )


class family_rent(Variable):
    value_type = float
    entity = BenUnit
    label = u"Gross rent for the family"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        personal_rent = benunit.members("personal_rent", period)
        return benunit.sum(personal_rent)


class childcare_expenses(Variable):
    value_type = float
    entity = Person
    label = u"Cost of childcare"
    definition_period = YEAR


class private_pension_contributions(Variable):
    value_type = float
    entity = Person
    label = u"Private pension contributions"
    definition_period = YEAR


class occupational_pension_contributions(Variable):
    value_type = float
    entity = Person
    label = u"Occupational pension contributions"
    definition_period = YEAR


class housing_service_charges(Variable):
    value_type = float
    entity = Household
    label = u"Housing service charges"
    definition_period = YEAR


class water_and_sewerage_charges(Variable):
    value_type = float
    entity = Household
    label = u"Water and sewerage charges"
    definition_period = YEAR


class employer_pension_contributions(Variable):
    value_type = float
    entity = Person
    label = u"Employer pension contributions"
    definition_period = YEAR


class weekly_childcare_expenses(Variable):
    value_type = float
    entity = Person
    label = u"Average cost of childcare"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("childcare_expenses", period) / WEEKS_IN_YEAR


@uprated(by="CPI")
class housing_costs(Variable):
    value_type = float
    entity = Household
    label = u"Total housing costs"
    definition_period = YEAR

    def formula(household, period, parameters):
        return household("rent", period) + household("mortgage", period)


class maintenance_expenses(Variable):
    value_type = float
    entity = Person
    label = u"Maintenance expenses"
    definition_period = YEAR


class mortgage_interest_repayment(Variable):
    value_type = float
    entity = Household
    label = u"Total mortgage payments"
    definition_period = YEAR


class mortgage_capital_repayment(Variable):
    value_type = float
    entity = Household
    label = u"Mortgage payments"
    definition_period = YEAR


class mortgage(Variable):
    value_type = float
    entity = Household
    label = u"Total mortgage payments"
    definition_period = YEAR


@uprated(by="council_tax")
class council_tax(Variable):
    value_type = float
    entity = Household
    label = u"Council Tax"
    definition_period = YEAR


class CouncilTaxBand(Enum):
    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"
    F = "F"
    G = "G"
    H = "H"
    I = "I"


class council_tax_band(Variable):
    value_type = Enum
    possible_values = CouncilTaxBand
    default_value = CouncilTaxBand.D
    entity = Household
    label = u"Council Tax Band"
    definition_period = YEAR



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *

"""
This file calculates the overall liability for Income Tax.
"""


class earned_taxable_income(Variable):
    value_type = float
    entity = Person
    label = u"Non-savings, non-dividend income for Income Tax"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 10"

    def formula(person, period, parameters):
        amount = person("adjusted_net_income", period) - add(
            person,
            period,
            ["taxable_savings_interest_income", "taxable_dividend_income"],
        )
        reductions = add(person, period, ["allowances", "marriage_allowance"])
        final_amount = max_(0, amount - reductions)
        return final_amount


class taxed_income(Variable):
    value_type = float
    entity = Person
    label = u"Income which is taxed"
    definition_period = YEAR

    def formula(person, period, parameters):
        COMPONENTS = [
            "earned_taxable_income",
            "taxed_savings_income",
            "taxed_dividend_income",
        ]
        return add(person, period, COMPONENTS)


class basic_rate_earned_income(Variable):
    value_type = float
    entity = Person
    label = u"Earned income (non-savings, non-dividend) at the basic rate"
    definition_period = YEAR

    def formula(person, period, parameters):
        income = person("earned_taxable_income", period)
        rates = parameters(period).tax.income_tax.rates
        amount = clip(income, rates.uk.thresholds[0], rates.uk.thresholds[1])
        return amount


class higher_rate_earned_income(Variable):
    value_type = float
    entity = Person
    label = u"Earned income (non-savings, non-dividend) at the higher rate"
    definition_period = YEAR

    def formula(person, period, parameters):
        income = person("earned_taxable_income", period)
        rates = parameters(period).tax.income_tax.rates
        amount = (
            clip(income, rates.uk.thresholds[1], rates.uk.thresholds[2])
            - rates.uk.thresholds[1]
        )
        return amount


class add_rate_earned_income(Variable):
    value_type = float
    entity = Person
    label = u"Earned income (non-savings, non-dividend) at the additional rate"
    definition_period = YEAR

    def formula(person, period, parameters):
        income = person("earned_taxable_income", period)
        rates = parameters(period).tax.income_tax.rates
        amount = (
            clip(income, rates.uk.thresholds[2], inf) - rates.uk.thresholds[2]
        )
        return amount


class basic_rate_earned_income_tax(Variable):
    value_type = float
    entity = Person
    label = u"Income tax on earned income at the basic rate"
    definition_period = YEAR

    def formula(person, period, parameters):
        amount = person("basic_rate_earned_income", period)
        charge = parameters(period).tax.income_tax.rates.uk.rates[0] * amount
        return charge


class higher_rate_earned_income_tax(Variable):
    value_type = float
    entity = Person
    label = u"Income tax on earned income at the higher rate"
    definition_period = YEAR

    def formula(person, period, parameters):
        amount = person("higher_rate_earned_income", period)
        charge = parameters(period).tax.income_tax.rates.uk.rates[1] * amount
        return charge


class add_rate_earned_income_tax(Variable):
    value_type = float
    entity = Person
    label = u"Income tax on earned income at the additional rate"
    definition_period = YEAR

    def formula(person, period, parameters):
        amount = person("add_rate_earned_income", period)
        charge = parameters(period).tax.income_tax.rates.uk.rates[2] * amount
        return charge


class earned_income_tax(Variable):
    value_type = float
    entity = Person
    label = u"Income tax on earned income"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 10"

    def formula(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        uk_amount = rates.uk.calc(person("earned_taxable_income", period))
        return uk_amount

    def formula_2017_04_06(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        scot = person("pays_scottish_income_tax", period)
        uk_amount = rates.uk.calc(person("earned_taxable_income", period))
        scot_amount = rates.scotland.pre_starter_rate.calc(
            person("earned_taxable_income", period)
        )
        amount = where(scot, scot_amount, uk_amount)
        return amount

    def formula_2018_04_06(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        scot = person("pays_scottish_income_tax", period)
        uk_amount = rates.uk.calc(person("earned_taxable_income", period))
        scot_amount = rates.scotland.post_starter_rate.calc(
            person("earned_taxable_income", period)
        )
        amount = where(scot, scot_amount, uk_amount)
        return amount


class TaxBand(Enum):
    NONE = "None"
    STARTER = "Starter (Scottish rates)"
    BASIC = "Basic"
    INTERMEDIATE = "Intermediate (Scottish rates)"
    HIGHER = "Higher"
    ADDITIONAL = "Additional"


class pays_scottish_income_tax(Variable):
    value_type = float
    entity = Person
    label = u"Whether the individual pays Scottish Income Tax rates"
    definition_period = YEAR

    def formula(person, period, parameters):
        country = person.household("country", period)
        countries = country.possible_values
        return person.household("country", period) == countries.SCOTLAND


class tax_band(Variable):
    value_type = Enum
    possible_values = TaxBand
    default_value = TaxBand.NONE
    entity = Person
    label = u"Tax band of the individual"
    definition_period = YEAR

    def formula(person, period, parameters):
        allowances = person("allowances", period)
        ANI = person("adjusted_net_income", period)
        rates = parameters(period).tax.income_tax.rates
        basic = allowances + rates.uk.thresholds[0]
        higher = allowances + rates.uk.thresholds[-2]
        add = allowances + rates.uk.thresholds[-1]
        band = select(
            [ANI >= add, ANI >= higher, ANI > basic, ANI <= basic],
            [TaxBand.ADDITIONAL, TaxBand.HIGHER, TaxBand.BASIC, TaxBand.NONE],
        )
        return band

    def formula_2017_04_06(person, period, parameters):
        allowances = person("allowances", period)
        ANI = person("adjusted_net_income", period)
        rates = parameters(period).tax.income_tax.rates
        scot = person("pays_scottish_income_tax", period)
        income = ANI - allowances
        uk_band = select(
            [income < threshold for threshold in rates.uk.thresholds] + [True],
            [TaxBand.NONE, TaxBand.BASIC, TaxBand.HIGHER, TaxBand.ADDITIONAL],
        )
        scottish_band = select(
            [
                income < threshold
                for threshold in rates.scotland.pre_starter_rate.thresholds
            ]
            + [True],
            [TaxBand.NONE, TaxBand.BASIC, TaxBand.HIGHER, TaxBand.ADDITIONAL],
        )
        return where(scot, scottish_band, uk_band)

    def formula_2018_06_01(person, period, parameters):
        allowances = person("allowances", period)
        ANI = person("adjusted_net_income", period)
        rates = parameters(period).tax.income_tax.rates
        scot = person("pays_scottish_income_tax", period)
        income = ANI - allowances
        uk_band = select(
            [income < threshold for threshold in rates.uk.thresholds[:3]]
            + [True],
            [TaxBand.NONE, TaxBand.BASIC, TaxBand.HIGHER, TaxBand.ADDITIONAL],
        )
        scottish_band = select(
            [
                income < threshold
                for threshold in rates.scotland.post_starter_rate.thresholds[
                    :5
                ]
            ]
            + [True],
            [
                TaxBand.NONE,
                TaxBand.STARTER,
                TaxBand.BASIC,
                TaxBand.INTERMEDIATE,
                TaxBand.HIGHER,
                TaxBand.ADDITIONAL,
            ],
        )
        return where(scot, scottish_band, uk_band)


class basic_rate_savings_income_pre_starter(Variable):
    value_type = float
    entity = Person
    label = u"Savings income which would otherwise be taxed at the basic rate, without the starter rate"
    definition_period = YEAR

    def formula(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates.uk
        savings_income_total = person(
            "taxable_savings_interest_income", period
        )
        savings_allowance = person("savings_allowance", period)
        savings_income = max_(0, savings_income_total - savings_allowance)
        other_income = person("earned_taxable_income", period)
        basic_rate_amount_with_savings = clip(
            savings_income + other_income,
            rates.thresholds[0],
            rates.thresholds[1],
        )
        basic_rate_amount_without_savings = clip(
            other_income, rates.thresholds[0], rates.thresholds[1]
        )
        amount = (
            basic_rate_amount_with_savings - basic_rate_amount_without_savings
        )
        return amount


class savings_starter_rate_income(Variable):
    value_type = float
    entity = Person
    label = u"Savings income which is tax-free under the starter rate"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 12"

    def formula(person, period, parameters):
        income = person("basic_rate_savings_income_pre_starter", period)
        limit = parameters(
            period
        ).tax.income_tax.rates.savings_starter_rate.allowance
        return max_(0, limit - income)


class basic_rate_savings_income(Variable):
    value_type = float
    entity = Person
    label = u"Savings income at the basic rate"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 11D"

    def formula(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        other_income = person("earned_taxable_income", period)
        savings_deductions = add(
            person,
            period,
            ["savings_allowance", "savings_starter_rate_income"],
        )
        savings_income_less_deductions = max_(
            0,
            person("taxable_savings_interest_income", period)
            - savings_deductions,
        )
        basic_rate_amount_with = clip(
            other_income + savings_income_less_deductions,
            rates.uk.thresholds[0],
            rates.uk.thresholds[1],
        )
        basic_rate_amount_without = clip(
            other_income, rates.uk.thresholds[0], rates.uk.thresholds[1]
        )
        basic_rate_amount = max_(
            0, basic_rate_amount_with - basic_rate_amount_without
        )
        return basic_rate_amount


class higher_rate_savings_income(Variable):
    value_type = float
    entity = Person
    label = u"Savings income at the higher rate"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 11D"

    def formula(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        other_income = person("earned_taxable_income", period)
        savings_deductions = add(
            person,
            period,
            ["savings_allowance", "savings_starter_rate_income"],
        )
        savings_income_less_deductions = max_(
            0,
            person("taxable_savings_interest_income", period)
            - savings_deductions,
        )
        higher_rate_amount_with = clip(
            other_income + savings_income_less_deductions,
            rates.uk.thresholds[1],
            rates.uk.thresholds[2],
        )
        higher_rate_amount_without = clip(
            other_income, rates.uk.thresholds[1], rates.uk.thresholds[2]
        )
        higher_rate_amount = max_(
            0, higher_rate_amount_with - higher_rate_amount_without
        )
        return higher_rate_amount


class add_rate_savings_income(Variable):
    value_type = float
    entity = Person
    label = u"Savings income at the higher rate"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 11D"

    def formula(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        other_income = person("earned_taxable_income", period)
        savings_deductions = add(
            person,
            period,
            ["savings_allowance", "savings_starter_rate_income"],
        )
        savings_income_less_deductions = max_(
            0,
            person("taxable_savings_interest_income", period)
            - savings_deductions,
        )
        add_rate_amount_with = clip(
            other_income + savings_income_less_deductions,
            rates.uk.thresholds[2],
            inf,
        )
        add_rate_amount_without = clip(
            other_income, rates.uk.thresholds[2], inf
        )
        add_rate_amount = max_(
            0, add_rate_amount_with - add_rate_amount_without
        )
        return add_rate_amount


class taxed_savings_income(Variable):
    value_type = float
    entity = Person
    label = u"Savings income which advances the person's income tax schedule"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 11D"

    def formula(person, period, parameters):
        COMPONENTS = [
            "basic_rate_savings_income",
            "higher_rate_savings_income",
            "add_rate_savings_income",
        ]
        amount = add(person, period, COMPONENTS)
        return amount


class taxed_dividend_income(Variable):
    value_type = float
    entity = Person
    label = u"Dividend income which is taxed"
    definition_period = YEAR

    def formula(person, period, parameters):
        return max_(
            0,
            person("taxable_dividend_income", period)
            - person("dividend_allowance", period),
        )


class savings_income_tax(Variable):
    value_type = float
    entity = Person
    label = u"Income tax on savings income"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 11D"

    def formula(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        basic_rate_amount = person("basic_rate_savings_income", period)
        higher_rate_amount = person("higher_rate_savings_income", period)
        add_rate_amount = person("add_rate_savings_income", period)
        charge = (
            rates.uk.rates[0] * basic_rate_amount
            + rates.uk.rates[1] * higher_rate_amount
            + rates.uk.rates[2] * add_rate_amount
        )
        return charge


class dividend_income_tax(Variable):
    value_type = float
    entity = Person
    label = u"Income tax on dividend income"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 13"

    def formula(person, period, parameters):
        rates = parameters(period).tax.income_tax.rates
        other_income = person("earned_taxable_income", period) + person(
            "taxed_savings_income", period
        )
        taxable_dividends = person("taxed_dividend_income", period)
        tax_with_dividends = rates.dividends.calc(
            other_income + taxable_dividends
        )
        tax_without_dividends = rates.dividends.calc(other_income)
        dividend_tax = tax_with_dividends - tax_without_dividends
        return dividend_tax


class income_tax_pre_charges(Variable):
    value_type = float
    entity = Person
    label = u"Income Tax before any tax charges"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 23"

    def formula(person, period, parameters):
        COMPONENTS = [
            "earned_income_tax",
            "savings_income_tax",
            "dividend_income_tax",
        ]
        total = add(person, period, COMPONENTS)
        return total


class is_higher_earner(Variable):
    value_type = bool
    entity = Person
    label = u"Whether this person is the highest earner in a family"
    definition_period = YEAR

    def formula(person, period, parameters):
        income = person("adjusted_net_income", period)
        # Add noise to incomes in order to avoid ties
        return (
            person.get_rank(person.benunit, -income + random(person) * 1e-2)
            == 0
        )


class CB_HITC(Variable):
    value_type = float
    entity = Person
    label = u"Child Benefit High-Income Tax Charge"
    definition_period = YEAR
    reference = "Finance Act 2012 s. 681B"

    def formula(person, period, parameters):
        CB_received = person.benunit("child_benefit", period, options=[ADD])
        CB_HITC = parameters(period).tax.income_tax.charges.CB_HITC
        phase_length = CB_HITC.phase_out_end - CB_HITC.phase_out_start
        income = amount_between(
            person("adjusted_net_income", period),
            CB_HITC.phase_out_start,
            CB_HITC.phase_out_end,
        )
        percentage = income / phase_length
        return (percentage * CB_received) * person("is_higher_earner", period)


class income_tax(Variable):
    value_type = float
    entity = Person
    label = "Income Tax"
    documentation = "Total Income Tax liability"
    definition_period = YEAR
    unit = "currency-GBP"
    reference = "Income Tax Act 2007 s. 23"

    def formula(person, period, parameters):
        return max_(
            0,
            add(person, period, ["income_tax_pre_charges", "CB_HITC"])
            - person("married_couples_allowance_deduction", period),
        )



from openfisca_uk.model_api import *


class child_benefit_reported(Variable):
    label = "Child Benefit (reported amount)"
    documentation = "Reported amount received for Child Benefit"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"


class would_claim_child_benefit(Variable):
    label = "Would claim Child Benefit"
    documentation = (
        "Whether this benefit unit would claim Child Benefit if eligible"
    )
    entity = BenUnit
    definition_period = YEAR
    value_type = bool

    def formula(benunit, period, parameters):
        claims_benefits = benunit("claims_all_entitled_benefits", period)
        takeup_rate = parameters(period).hmrc.child_benefit.takeup_rate
        imputed_takeup = (
            benunit("benunit_random_number", period) <= takeup_rate
        )
        return claims_benefits | imputed_takeup


class child_benefit_respective_amount(Variable):
    label = "Child Benefit (respective amount)"
    documentation = "The amount of this benefit unit's Child Benefit which is in respect of this person"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"
    reference = (
        "https://www.legislation.gov.uk/ukpga/1992/4/part/IX",
        "https://www.legislation.gov.uk/uksi/2006/965/regulation/2",
    )

    def formula(person, period, parameters):
        eligible = person("is_child_or_QYP", period)
        is_eldest = person("is_eldest_child", period)
        child_benefit = parameters(period).hmrc.child_benefit.amount
        amount = where(
            is_eldest, child_benefit.eldest, child_benefit.additional
        )
        return eligible * amount * WEEKS_IN_YEAR


class child_benefit(Variable):
    label = "Child Benefit"
    documentation = "Total Child Benefit for the benefit unit"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(benunit, period):
        entitlement = benunit.sum(
            benunit.members("child_benefit_respective_amount", period)
        )
        return entitlement * benunit("would_claim_child_benefit", period)


class child_benefit_less_tax_charge(Variable):
    label = "Child Benefit (less tax charge)"
    documentation = (
        "Child Benefit, minus the Child Benefit High-Income Tax Charge"
    )
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(benunit, period):
        benefit = benunit("child_benefit", period)
        charge = benunit.sum(benunit.members("CB_HITC", period))
        return benefit - charge



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

setup(
    name="OpenFisca-UK",
    version="0.7.12",
    author="PolicyEngine",
    author_email="nikhil@policyengine.org",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    description="OpenFisca tax and benefit system for UK",
    keywords="benefit microsimulation social tax",
    license="http://www.fsf.org/licensing/licenses/agpl-3.0.html",
    url="https://github.com/pslmodels/openfisca-uk",
    include_package_data=True,  # Will read MANIFEST.in
    data_files=[
        (
            "share/openfisca/openfisca-country-template",
            ["CHANGELOG.md", "LICENSE", "README.md"],
        ),
    ],
    install_requires=[
        "requests>=2.25.1",
        "pandas>=1.1.4",
        "numpy>=1.17.5",
        "pyyaml>=5.3.1",
        "pytest>=5.4.3",
        "OpenFisca-Core>=35.4.1",
        "microdf_python>=0.3.0",
        "OpenFisca-UK-Data>=0.5.3",
        "OpenFisca-Tools>=0.1.7,<0.2.0",
        "tqdm>=4.59.0",
        "plotly>=4.14.3",
        "argparse>=1.4.0",
        "inquirer>=2.7.0",
    ],
    extras_require={
        "dev": [
            "autopep8 >=1.5",
            "flake8 >=3.5.0,<3.8.0",
            "flake8-print",
            "pycodestyle >=2.3.0,<2.6.0",  # To avoid incompatibility with flake
        ]
    },
    entry_points={
        "console_scripts": [
            "openfisca-uk-setup=openfisca_uk.initial_setup:main"
        ],
    },
    packages=find_packages(),
)



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class PIP(Variable):
    value_type = float
    entity = Person
    label = u"Personal Independence Payment"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("PIP_M", period) + person("PIP_DL", period)


class PIP_DL(Variable):
    value_type = float
    entity = Person
    label = u"Personal Independence Payment (Daily Living)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("PIP_DL_reported", period)


class PIP_M(Variable):
    value_type = float
    entity = Person
    label = u"Personal Independence Payment (Mobility)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("PIP_M_reported", period)


class PIP_DL_reported(Variable):
    value_type = float
    entity = Person
    label = u"Personal Independence Payment (Daily Living) (reported)"
    definition_period = YEAR


class PIP_M_reported(Variable):
    value_type = float
    entity = Person
    label = u"Personal Independence Payment (Mobility) (reported)"
    definition_period = YEAR



from openfisca_uk.tools.general import *
from openfisca_uk.entities import *


class state_pension_age(Variable):
    value_type = float
    entity = Person
    label = u"State Pension age for this person"
    definition_period = YEAR

    def formula(person, period, parameters):
        SP = parameters(period).benefit.state_pension
        male = person("is_male", period)
        threshold = male * SP.male_age + not_(male) * SP.female_age
        return threshold


class is_SP_age(Variable):
    value_type = bool
    entity = Person
    label = u"Whether the person is State Pension Age"
    definition_period = YEAR

    def formula(person, period, parameters):
        age = person("age", period)
        threshold = person("state_pension_age", period)
        over = age >= threshold
        return over


class triple_lock_uprating(Variable):
    value_type = float
    entity = Person
    label = u"Triple lock relative increase"
    documentation = (
        "A government commitment, rather than a legislative requirement"
    )
    definition_period = YEAR

    def formula(person, period, parameters):
        cpi_growth = (
            parameters(period).uprating.CPI
            / parameters(period.last_year).uprating.CPI
        )
        earnings_growth = (
            parameters(period).uprating.earnings
            / parameters(period.last_year).uprating.earnings
        )
        return max(
            parameters(period).benefit.state_pension.triple_lock_minimum,
            cpi_growth,
            earnings_growth,
        )


class state_pension(Variable):
    value_type = float
    entity = Person
    label = u"Income from the State Pension"
    definition_period = YEAR
    unit = "currency-GBP"
    documentation = "Gross State Pension payments"

    def formula(person, period, parameters):
        return person("state_pension_reported", period)


class state_pension_reported(Variable):
    value_type = float
    entity = Person
    label = u"Reported income from the State Pension"
    definition_period = YEAR

    def formula_2015(person, period, parameters):
        return person("state_pension_reported", period.last_year) * person(
            "triple_lock_uprating", period
        )
